from queue import Queue
from threading import Event

import cv2
import numpy as np
from PySide6.QtCore import Qt, QThread, Signal
from PySide6.QtGui import QPixmap

import caliscope.calibration.draw_charuco as draw_charuco
import caliscope.logger
from caliscope.gui.frame_emitters.tools import apply_rotation, cv2_to_qlabel, resize_to_square
from caliscope.recording.recorded_stream import RecordedStream

logger = caliscope.logger.get(__name__)


class PlaybackFrameEmitter(QThread):
    # establish signals that will be displayed within the GUI
    ImageBroadcast = Signal(int, QPixmap)
    GridCountBroadcast = Signal(int)
    FrameIndexBroadcast = Signal(int, int)

    def __init__(self, recorded_stream: RecordedStream, grid_history_q: Queue, pixmap_edge_length=500):
        # pixmap_edge length is from the display window. Keep the display area
        # square to keep life simple.
        super(PlaybackFrameEmitter, self).__init__()
        self.stream = recorded_stream
        self.port = self.stream.port

        # Apply a safety margin to the scaling factors (e.g., 5%)
        # used only when applying the undistortion
        self.scaling_factor = 1

        self.frame_packet_q = Queue()
        self.grid_history_q = grid_history_q  # received a tuple of ids, img_loc

        self.stream.subscribe(self.frame_packet_q)
        self.pixmap_edge_length = pixmap_edge_length
        self.undistort = False
        self.keep_collecting = Event()
        self.initialize_grid_capture_history()

    def initialize_grid_capture_history(self):
        """
        The grid capture history is only used as a GUI element to display to the user
        the corners that have been collected and are being used in the calibration.

        It is not otherwise used in any calculations and needs to be re-initialized by
        the intrinsic stream manager whenever the intrinsic calibrators data is also
        re-initialized.
        """
        self.connected_points = self.stream.tracker.get_connected_points()
        width = self.stream.size[0]
        height = self.stream.size[1]
        channels = 3
        self.grid_capture_history = np.zeros((height, width, channels), dtype="uint8")

    def run(self):
        self.keep_collecting.set()

        while self.keep_collecting.is_set():
            # Grab a frame from the queue and broadcast to displays
            # self.monocalibrator.grid_frame_ready_q.get()
            logger.debug("Getting frame packet from queue")
            frame_packet = self.frame_packet_q.get()

            while self.grid_history_q.qsize() > 0:
                ids, img_loc = self.grid_history_q.get()
                self.add_to_grid_history(ids, img_loc)

            if not self.keep_collecting.is_set():
                break

            if frame_packet.frame is not None:  # stream end signal when None frame placed on out queue
                self.frame = frame_packet.frame_with_points

                logger.debug(f"Frame size is {self.frame.shape}")
                logger.debug(f"Grid Capture History size is {self.grid_capture_history.shape}")
                self.frame = cv2.addWeighted(self.frame, 1, self.grid_capture_history, 1, 0)

                self._apply_undistortion()

                logger.debug(f"Frame size is {self.frame.shape} following undistortion")
                self.frame = resize_to_square(self.frame)
                self.frame = apply_rotation(self.frame, self.stream.rotation_count)
                image = cv2_to_qlabel(self.frame)
                pixmap = QPixmap.fromImage(image)

                if self.pixmap_edge_length:
                    pixmap = pixmap.scaled(
                        int(self.pixmap_edge_length),
                        int(self.pixmap_edge_length),
                        Qt.AspectRatioMode.KeepAspectRatio,
                    )
                self.ImageBroadcast.emit(self.port, pixmap)
                self.FrameIndexBroadcast.emit(self.port, frame_packet.frame_index)

        logger.info(f"Thread loop within frame emitter at port {self.stream.port} successfully ended")

    def stop(self):
        logger.info(f"Beginning to shut down frame emitter at port {self.port}")
        self.stream.unsubscribe(self.frame_packet_q)
        self.keep_collecting.clear()
        self.frame_packet_q.put(-1)
        self.quit()

    def set_scale_factor(self, scaling_factor):
        self.scaling_factor = scaling_factor

        if hasattr(self, "matrix") and hasattr(self, "distortions"):
            self.update_distortion_params(self.undistort, self.matrix, self.distortions)

    def update_distortion_params(self, undistort=None, matrix=None, distortions=None):
        if matrix is None:
            logger.info(f"No camera matrix calculated yet at port {self.port}")
        else:
            logger.info(f"Updating camera matrix and distortion parameters for frame emitter at port {self.port}")

            self.undistort = undistort
            self.matrix = matrix
            self.distortions = distortions

            h, w = self.stream.size
            # h, w = original_image_size
            initial_new_matrix, valid_roi = cv2.getOptimalNewCameraMatrix(self.matrix, self.distortions, (w, h), 1)

            logger.info(f"Valid ROI is {valid_roi}")

            # Find extreme points and midpoints in the original image
            corners = np.array([[0, 0], [0, h], [w, 0], [w, h]], dtype=np.float32)
            midpoints = np.array([[w / 2, 0], [w / 2, h], [0, h / 2], [w, h / 2]], dtype=np.float32)
            extreme_points = np.vstack((corners, midpoints))

            # Undistort these points
            undistorted_points = cv2.undistortPoints(
                np.expand_dims(extreme_points, axis=1),
                self.matrix,
                self.distortions,
                P=initial_new_matrix,
            )

            # Find min/max x and y in undistorted points
            min_x = min(undistorted_points[:, 0, 0])
            max_x = max(undistorted_points[:, 0, 0])
            min_y = min(undistorted_points[:, 0, 1])
            max_y = max(undistorted_points[:, 0, 1])

            # # Calculate new image width and height
            # new_width = int(np.ceil(max_x - min_x))
            # new_height = int(np.ceil(max_y - min_y))

            # Calculate scaling factors
            scale_x = (max_x - min_x) / w
            scale_y = (max_y - min_y) / h

            scale_x *= self.scaling_factor
            scale_y *= self.scaling_factor

            # Adjust new image width and height
            adjusted_width = int(w * scale_x)
            adjusted_height = int(h * scale_y)

            logger.info(f"New image size for undistorted frame: {(adjusted_width,adjusted_height)}")
            # Now use new_width and new_height as your NewImageSize for undistortion
            # newImageSize = (new_width, new_height)
            self.new_matrix, valid_roi = cv2.getOptimalNewCameraMatrix(
                self.matrix, self.distortions, (w, h), 1, (adjusted_width, adjusted_height)
            )

    def _apply_undistortion(self):
        if self.undistort and self.matrix is not None:
            # Compute the optimal new camera matrix
            # Undistort the image
            self.frame = cv2.undistort(self.frame, self.matrix, self.distortions, None, self.new_matrix)

    def add_to_grid_history(self, ids, img_loc):
        """
        Note that the connected points here comes from the charuco tracker.
        This grid history is likely best tracked by the controller and
        a reference should be past to the frame emitter
        """
        # logger.info("Attempting to add to grid history")
        if len(ids) > 3:
            # logger.info("enough points to add")
            self.grid_capture_history = draw_charuco.grid_history(
                self.grid_capture_history,
                ids,
                img_loc,
                self.connected_points,
            )
        else:
            logger.info("Not enough points....grid not added...")



from abc import ABC, abstractmethod
from dataclasses import dataclass

import cv2
import numpy as np
from numba.typed import List


@dataclass(frozen=True, slots=True)
class PointPacket:
    """
    This will be the primary return value of the Tracker

    Note that obj_loc will generally be `None`. These are the point positions in
    the object's frame of reference.
    It has actual values when using the Charuco tracker as these are used in the calibration.
    """

    point_id: np.ndarray = None  # unique point id that aligns with Tracker.get_point_names()
    img_loc: np.ndarray = None  # x,y position of tracked point
    obj_loc: np.ndarray = None  # x,y,z in object frame of reference; primarily for calibration
    confidence: np.ndarray = None  # may be available in some trackers..include for potentnial downstream calculations

    @property
    def obj_loc_list(self) -> List[List]:
        """
        if obj loc is none, printing data to a table format will be undermined
        this creates a list of length len(point_id) that is empty so that a pd.dataframe
        can be constructed from it.
        """
        if self.obj_loc is not None:
            obj_loc_x = self.obj_loc[:, 0].tolist()
            obj_loc_y = self.obj_loc[:, 1].tolist()
        else:
            length = len(self.point_id)
            obj_loc_x = [None] * length
            obj_loc_y = [None] * length

        return [obj_loc_x, obj_loc_y]


class Tracker(ABC):
    @property
    def name(self) -> str:
        """
        returns the tracker name
        This name should align with the label used by TrackerEnum
        Used for file naming creation
        """
        pass

    @abstractmethod
    def get_points(self, frame: np.ndarray, port: int, rotation_count: int) -> PointPacket:
        """
        frame: np.ndarray from reading an OpenCV capture object

        port: integer value usd to track which camera the frame originates from

        rotation count: used to indicate the orientation of the image (e.g. rotateed 90 degrees left or right)
                        Some tracking algorithms expect images to be "upright", so this can be used to align the image

                        The function `apply_rotation` from `caliscope.trackers.helper` can correctly orient the image
                        The function `unrotate_points` from the same module can convert any tracked points back into
                        the original orientation
        """
        pass

    @abstractmethod
    def get_point_name(self, point_id: int) -> str:
        """
        Maps point_id to a name
        Used for saving out data with sensible headers.
        """
        pass

    @abstractmethod
    def get_point_id(self, point_name: str) -> int:
        """
        Maps point name to point_id
        """
        pass

    @abstractmethod
    def scatter_draw_instructions(self, point_id: int) -> dict:
        """
        Maps point_id to a dictionary of parameters used to draw circles on frames for visual feedback.

        As an example, the dictionary could have the form: {"radius": 5, "color": (220, 0, 0), "thickness": 3}
        The parameters `radius`, `color`, and `thickness` are used downstream in a call to `cv2.circle`
        to place the tracked point on the frame. See `FramePacket` below.
        """
        pass

    def get_connected_points(self) -> dict[str : tuple[int, int, tuple[int, int]]]:
        """
        OPTIONAL METHOD
        used for drawing purposes elsewhere. Specify which
        points (if any) should have a line connecting them
        {SegmentName:(pointNameA, pointNameB, )}
        """
        pass

    @property
    def metarig_mapped(self):
        """
        OPTIONAL PROPERTY

        Defaults to false and can be overriden to True
        Used to ensure that metarig_config creation is not presented as
        an option in GUI
        """
        return False

    @property
    def metarig_symmetrical_measures(self):
        """
        OPTIONAL PROPERTY

        a dictionary of key: value in the form Measure_Name:[pointA, pointB]
        when processed, the mean distances (excluding outliers) of both
        left_pointA,left_pointB and right_pointA, right_pointB will be calculated.
        The mean of the two sides will be taken
        """
        raise NotImplementedError(f"Tracker {self.name} has not provided its measures for configuring a metarig")

    @property
    def metarig_bilateral_measures(self):
        """
        OPTIONAL PROPERTY

        a dictionary of key: value in the form Measure_Name:[side_pointA, side_pointB]
        when processed, mean distance (excluding outliers)
        between the two points will be calculated and stored as the measure
        """
        raise NotImplementedError(f"Tracker {self.name} has not provided its measures for configuring a metarig")


@dataclass(frozen=True, slots=True)
class FramePacket:
    """
    Holds the data for a single frame from a camera, including the frame itself,
    the frame time and the points if they were generated
    """

    port: int
    frame_index: int
    frame_time: float
    frame: np.ndarray
    points: PointPacket = None
    draw_instructions: callable = None

    def to_tidy_table(self, sync_index) -> dict:
        """
        Returns a dictionary of lists where each list is as long as the
        number of points identified on the frame;
        used for creating csv output via pandas
        """
        if self.points is not None:
            point_count = len(self.points.point_id)
            if point_count > 0:
                table = {
                    "sync_index": [sync_index] * point_count,
                    "port": [self.port] * point_count,
                    "frame_index": [self.frame_index] * point_count,
                    "frame_time": [self.frame_time] * point_count,
                    "point_id": self.points.point_id.tolist(),
                    "img_loc_x": self.points.img_loc[:, 0].tolist(),
                    "img_loc_y": self.points.img_loc[:, 1].tolist(),
                    "obj_loc_x": self.points.obj_loc_list[0],
                    "obj_loc_y": self.points.obj_loc_list[1],
                }
            else:
                table = None
        else:
            table = None
        return table

    @property
    def frame_with_points(self):
        if self.points is not None:
            drawn_frame = self.frame.copy()
            ids = self.points.point_id
            locs = self.points.img_loc
            for _id, coord in zip(ids, locs):
                x = round(coord[0])
                y = round(coord[1])

                # draw instructions are a method of Tracker object
                params = self.draw_instructions(_id)
                cv2.circle(
                    drawn_frame,
                    (x, y),
                    params["radius"],
                    params["color"],
                    params["thickness"],
                )
        else:
            drawn_frame = self.frame

        return drawn_frame


@dataclass(frozen=True, slots=True)
class SyncPacket:
    """
    SyncPacket holds syncronized frame packets.
    """

    sync_index: int
    frame_packets: dict

    @property
    def triangulation_inputs(self):
        """
        returns three key items used by the triangulation functions
            cameras: a list of the camera ids associated with each reported 2d point
            point_ids: the point id associated with each 2d point
            img_xy: the 2d image points themselves

        """
        cameras = []
        point_ids = []
        img_xy = []

        for port, packet in self.frame_packets.items():
            if packet is not None and packet.points is not None:
                cameras.extend([port] * len(packet.points.point_id))
                point_ids.extend(packet.points.point_id.tolist())
                img_xy.extend(packet.points.img_loc.tolist())

        return cameras, point_ids, img_xy

    @property
    def dropped(self):
        """
        convencience method to ease tracking of dropped frame rate within the synchronizer
        """
        temp_dict = {}
        for port, packet in self.frame_packets.items():
            if packet is None:
                temp_dict[port] = 1
            else:
                temp_dict[port] = 0
        return temp_dict

    @property
    def frame_packet_count(self):
        count = 0
        for port, packet in self.frame_packets.items():
            if packet is not None:
                count += 1
        return count


@dataclass(slots=True, frozen=True)
class XYZPacket:
    sync_index: int
    point_ids: np.ndarray  # (n,1)
    point_xyz: np.ndarray  # (n,3)

    def get_point_xyz(self, point_id: int) -> np.ndarray:
        return self.point_xyz[self.point_ids == point_id]

    def get_segment_ends(self, point_id_A: int, point_id_B: int) -> np.ndarray:
        return np.vstack([self.get_point_xyz(point_id_A), self.get_point_xyz(point_id_B)])



import logging
from dataclasses import dataclass
from itertools import combinations
from pathlib import Path

import numpy as np

import caliscope.logger
from caliscope.cameras.synchronizer import Synchronizer
from caliscope.packets import SyncPacket

logger = caliscope.logger.get(__name__)


class StereoPointsBuilder:
    def __init__(self, ports: list):
        self.ports = ports
        self.pairs = [(i, j) for i, j in combinations(self.ports, 2) if i < j]

    def _get_stereo_points_packet(self, sync_index, port_A, points_A, port_B, points_B):
        # get ids in common
        if len(points_A.point_id) > 0 and len(points_B.point_id) > 0:
            common_ids = np.intersect1d(points_A.point_id, points_B.point_id)
        else:
            common_ids = np.array([])

        if len(common_ids) == 0:
            packet = None
        else:
            # common_ids = common_ids[:,0]
            # for both ports, get the indices of the common ids
            sorter_A = np.argsort(points_A.point_id)
            shared_indices_A = sorter_A[np.searchsorted(points_A.point_id, common_ids, sorter=sorter_A)]
            shared_indices_A

            sorter_B = np.argsort(points_B.point_id)
            shared_indices_B = sorter_B[np.searchsorted(points_B.point_id, common_ids, sorter=sorter_B)]
            shared_indices_B

            packet = StereoPointsPacket(
                sync_index=sync_index,
                port_A=port_A,
                port_B=port_B,
                common_ids=common_ids,
                img_loc_A=points_A.img_loc[shared_indices_A],
                img_loc_B=points_B.img_loc[shared_indices_B],
            )

            logger.debug(f"Points in common for ports ({port_A}, {port_B}): {common_ids}")

        return packet

    def get_synched_paired_points(self, sync_packet: SyncPacket):
        # will be populated with dataframes of:
        # id | img_x | img_y | board_x | board_y
        sync_index = sync_packet.sync_index
        paired_points_packets = {}

        for pair in self.pairs:
            port_A = pair[0]
            port_B = pair[1]

            if sync_packet.frame_packets[port_A] is not None and sync_packet.frame_packets[port_B] is not None:
                points_A = sync_packet.frame_packets[port_A].points
                points_B = sync_packet.frame_packets[port_B].points

                paired_points: StereoPointsPacket = self._get_stereo_points_packet(
                    sync_index, port_A, points_A, port_B, points_B
                )

                paired_points_packets[pair] = paired_points

        return SynchedStereoPointsPacket(sync_index, paired_points_packets)


@dataclass
class StereoPointsPacket:
    """The points shared by two FramePointsPackets"""

    sync_index: int

    port_A: int
    port_B: int

    common_ids: np.ndarray
    img_loc_A: np.ndarray
    img_loc_B: np.ndarray

    # a place to hold the pairwise triangulated value down the line
    xyz: np.ndarray = None

    @property
    def pair(self):
        return (self.port_A, self.port_B)

    def to_table(self):
        # table will be in the form of a dictionary of lists of equal length
        table = {}

        point_count = len(self.common_ids)

        table["pair"] = [self.pair] * point_count
        table["port_A"] = [self.port_A] * point_count
        table["port_B"] = [self.port_B] * point_count
        table["sync_index"] = [self.sync_index] * point_count
        table["point_id"] = list(self.common_ids)
        table["x_pos"] = list(self.xyz[:, 0])
        table["y_pos"] = list(self.xyz[:, 1])
        table["z_pos"] = list(self.xyz[:, 2])
        table["x_A"] = list(self.img_loc_A[:, 0])
        table["y_A"] = list(self.img_loc_A[:, 1])
        table["x_B"] = list(self.img_loc_B[:, 0])
        table["y_B"] = list(self.img_loc_B[:, 1])

        return table


@dataclass
class SynchedStereoPointsPacket:
    sync_index: int
    stereo_points_packets: dict

    @property
    def pairs(self):
        return list(self.stereo_points_packets.keys())

    def to_table(self):
        pass


if __name__ == "__main__":
    from caliscope.calibration.charuco import Charuco
    from caliscope.recording.recorded_stream import RecordedStreamPool

    logger.setLevel(logging.DEBUG)
    from caliscope import __root__

    session_directory = Path(__root__, "tests", "5_cameras", "recording")
    csv_output = Path(session_directory, "paired_point_data.csv")

    ports = [0, 1, 2, 3, 4]

    tracker = Charuco(4, 5, 11, 8.5, aruco_scale=0.75, square_size_overide_cm=5.25, inverted=True)

    recorded_stream_pool = RecordedStreamPool(ports, session_directory, charuco=tracker)
    syncr = Synchronizer(recorded_stream_pool.streams, fps_target=200)
    recorded_stream_pool.play_videos()

    point_stream = StereoPointsBuilder(synchronizer=syncr, csv_output_path=csv_output)

    # I think that EOF needs to propogate up
    while not point_stream.frames_complete:
        points_packet = point_stream.out_q.get()

        # print("--------------------------------------")
        # print(points_packet)

    print("Saving data....")
    # save_data = pd.DataFrame(point_stream.tidy_output)
    # save_data.to_csv(csv_output)



from queue import Queue
from threading import Thread

import cv2
import mediapipe as mp
import numpy as np

import caliscope.logger

# cap = cv2.VideoCapture(0)
from caliscope.packets import PointPacket
from caliscope.tracker import Tracker
from caliscope.trackers.helper import apply_rotation, unrotate_points

logger = caliscope.logger.get(__name__)


class HandTracker(Tracker):
    # Initialize MediaPipe Hands and Drawing utility
    def __init__(self) -> None:
        self.in_queue = Queue(-1)
        self.out_queue = Queue(-1)

        # each port gets its own mediapipe context manager
        # use a dictionary of queues for passing
        self.in_queues = {}
        self.out_queues = {}
        self.threads = {}

    @property
    def name(self):
        return "HAND"

    def run_frame_processor(self, port: int, rotation_count: int):
        # Create a MediaPipe Hands instance
        with mp.solutions.hands.Hands(
            static_image_mode=False,
            max_num_hands=2,
            min_detection_confidence=0.8,
            min_tracking_confidence=0.8,
        ) as hands:
            while True:
                frame = self.in_queues[port].get()
                # apply rotation as needed
                frame = apply_rotation(frame, rotation_count)

                height, width, color = frame.shape
                # Convert the image to RGB format
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                results = hands.process(frame)

                # initialize variables so none will be created if no points detected
                point_ids = []
                landmark_xy = []

                if results.multi_hand_landmarks:
                    # need to track left/right...more difficult than you might think
                    hand_types = []
                    for item in results.multi_handedness:
                        hand_info = item.ListFields()[0][1].pop()
                        hand_types.append(hand_info.label)

                    hand_type_index = 0

                    for hand_landmarks in results.multi_hand_landmarks:
                        # create adjusting factor to distinguish left/right
                        hand_label = hand_types[hand_type_index]
                        if hand_label == "Left":
                            side_adjustment_factor = 0
                        else:
                            side_adjustment_factor = 100

                        for landmark_id, landmark in enumerate(hand_landmarks.landmark):
                            point_ids.append(landmark_id + side_adjustment_factor)

                            # mediapipe expresses in terms of percent of frame, so must map to pixel position
                            x, y = int(landmark.x * width), int(landmark.y * height)
                            landmark_xy.append((x, y))

                        hand_type_index += 1

                point_ids = np.array(point_ids)
                landmark_xy = np.array(landmark_xy)
                landmark_xy = unrotate_points(landmark_xy, rotation_count, width, height)

                point_packet = PointPacket(point_ids, landmark_xy)

                self.out_queues[port].put(point_packet)

    def get_points(self, frame: np.ndarray, port: int, rotation_count: int) -> PointPacket:
        if port not in self.in_queues.keys():
            self.in_queues[port] = Queue(1)
            self.out_queues[port] = Queue(1)

            self.threads[port] = Thread(
                target=self.run_frame_processor,
                args=(port, rotation_count),
                daemon=True,
            )
            self.threads[port].start()

        self.in_queues[port].put(frame)
        point_packet = self.out_queues[port].get()

        return point_packet

    def get_point_name(self, point_id: int) -> str:
        return str(point_id)

    def scatter_draw_instructions(self, point_id: int) -> dict:
        if point_id < 100:
            rules = {"radius": 5, "color": (0, 0, 220), "thickness": 3}
        else:
            rules = {"radius": 5, "color": (220, 0, 0), "thickness": 3}
        return rules



import math
from threading import Event
from time import sleep

import cv2
import numpy as np
from PySide6.QtCore import QThread, Signal, Slot
from PySide6.QtGui import QImage, QPixmap
from PySide6.QtWidgets import (
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QWidget,
)

import caliscope.logger
from caliscope.cameras.synchronizer import Synchronizer
from caliscope.controller import Controller
from caliscope.packets import FramePacket

logger = caliscope.logger.get(__name__)


class ExtrinsicPlaybackWidget(QWidget):
    def __init__(self, controller: Controller):
        super(ExtrinsicPlaybackWidget, self).__init__()
        self.controller = controller
        self.synchronizer: Synchronizer = self.controller.synchronizer
        self.ports = self.synchronizer.ports

        # need to let synchronizer spin up before able to display frames
        while not hasattr(controller.synchronizer, "current_sync_packet"):
            sleep(0.25)
        # create tools to build and emit the displayed frame
        # self.unpaired_frame_builder = FramePrepper(self.synchronizer)
        self.thumbnail_emitter = FrameDictionaryEmitter(self.synchronizer)
        self.thumbnail_emitter.start()

        # all video output routed to qlabels stored in a dictionariy
        # make it as square as you can get it
        self.playback_displays = {str(port): QLabel() for port in self.ports}
        # self.recording_frame_display = QLabel()

        self.place_widgets()
        self.connect_widgets()

    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        self.settings_group = QGroupBox("Settings")
        self.settings_group.setLayout(QHBoxLayout())
        self.settings_group.layout().addWidget(QLabel("Frame Rate:"))
        self.settings_group.layout().addWidget(self.frame_rate_spin)
        self.layout().addWidget(self.settings_group)

        self.record_controls = QGroupBox()
        self.record_controls.setLayout(QHBoxLayout())
        self.record_controls.layout().addWidget(self.start_stop)
        self.record_controls.layout().addWidget(self.destination_label)
        self.record_controls.layout().addWidget(self.recording_directory)

        self.layout().addWidget(self.record_controls)

        dropped_fps_layout = QHBoxLayout()
        dropped_fps_layout.addStretch(1)
        dropped_fps_layout.addWidget(self.dropped_fps_label)
        dropped_fps_layout.addStretch(1)

        self.layout().addLayout(dropped_fps_layout)

        camera_count = len(self.ports)
        grid_columns = int(math.ceil(camera_count**0.5))
        # grid_rows = int(math.ceil(camera_count / grid_columns))

        frame_grid = QGridLayout()
        row = 0
        column = 0
        for port in sorted(self.ports):
            frame_grid.addWidget(self.playback_displays[str(port)], row, column)

            # update row and column for next iteration
            if column >= grid_columns - 1:
                # start fresh on next row
                column = 0
                row += 1
            else:
                column += 1

        frame_display_layout = QHBoxLayout()
        frame_display_layout.addStretch(1)
        frame_display_layout.addLayout(frame_grid)
        frame_display_layout.addStretch(1)
        self.layout().addLayout(frame_display_layout)

    def connect_widgets(self):
        self.thumbnail_emitter.ThumbnailImagesBroadcast.connect(self.ImageUpdateSlot)
        self.frame_rate_spin.valueChanged.connect(self.controller.set_active_mode_fps)
        self.thumbnail_emitter.dropped_fps.connect(self.update_dropped_fps)
        self.start_stop.clicked.connect(self.toggle_start_stop)
        self.controller.qt_signaler.recording_complete_signal.connect(self.on_recording_complete)

    @Slot(dict)
    def ImageUpdateSlot(self, q_image_dict: dict):
        logger.debug("About to get qpixmap from qimage")
        for port, thumbnail in q_image_dict.items():
            qpixmap = QPixmap.fromImage(thumbnail)
            logger.debug("About to set qpixmap to display")
            self.playback_displays[port].setPixmap(qpixmap)
            logger.debug("successfully set display")


class FrameDictionaryEmitter(QThread):
    ThumbnailImagesBroadcast = Signal(dict)
    dropped_fps = Signal(dict)

    def __init__(self, synchronizer: Synchronizer, single_frame_height=300):
        super(FrameDictionaryEmitter, self).__init__()
        self.single_frame_height = single_frame_height
        self.synchronizer = synchronizer
        logger.info("Initiated recording frame emitter")
        self.keep_collecting = Event()

    def run(self):
        self.keep_collecting.set()

        while self.keep_collecting.is_set():
            logger.debug("About to get next recording frame")
            # recording_frame = self.unpaired_frame_builder.get_recording_frame()

            logger.debug("Referencing current sync packet in synchronizer")
            self.current_sync_packet = self.synchronizer.current_sync_packet

            thumbnail_qimage = {}
            for port in self.synchronizer.ports:
                frame_packet = self.current_sync_packet.frame_packets[port]
                rotation_count = self.synchronizer.streams[port].camera.rotation_count

                text_frame = frame_packet_2_thumbnail(frame_packet, rotation_count, self.single_frame_height, port)
                q_image = cv2_to_qimage(text_frame)
                thumbnail_qimage[str(port)] = q_image

            self.ThumbnailImagesBroadcast.emit(thumbnail_qimage)

            dropped_fps_dict = {str(port): dropped for port, dropped in self.synchronizer.dropped_fps.items()}
            self.dropped_fps.emit(dropped_fps_dict)
        logger.info("Recording thumbnail emitter run thread ended...")


def frame_packet_2_thumbnail(frame_packet: FramePacket, rotation_count: int, edge_length: int, port: int):
    raw_frame = get_frame_or_blank(frame_packet, edge_length)
    # port = frame_packet.port

    # raw_frame = self.get_frame_or_blank(None)
    square_frame = resize_to_square(raw_frame, edge_length)
    rotated_frame = apply_rotation(square_frame, rotation_count)
    flipped_frame = cv2.flip(rotated_frame, 1)

    # put the port number on the top of the frame
    text_frame = cv2.putText(
        flipped_frame,
        str(port),
        (int(flipped_frame.shape[1] / 2), int(edge_length / 4)),
        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
        fontScale=1,
        color=(0, 0, 255),
        thickness=2,
    )

    return text_frame


def get_frame_or_blank(frame_packet: FramePacket, edge_length):
    """Synchronization issues can lead to some frames being None among
    the synched frames, so plug that with a blank frame"""

    if frame_packet is None:
        logger.debug("plugging blank frame data")
        frame = np.zeros((edge_length, edge_length, 3), dtype=np.uint8)
    else:
        frame = frame_packet.frame_with_points

    return frame


def resize_to_square(frame, edge_length):
    """To make sure that frames align well, scale them all to thumbnails
    squares with black borders."""
    logger.debug("resizing square")

    height = frame.shape[0]
    width = frame.shape[1]

    padded_size = max(height, width)

    height_pad = int((padded_size - height) / 2)
    width_pad = int((padded_size - width) / 2)
    pad_color = [0, 0, 0]

    logger.debug("about to pad border")
    frame = cv2.copyMakeBorder(
        frame,
        height_pad,
        height_pad,
        width_pad,
        width_pad,
        cv2.BORDER_CONSTANT,
        value=pad_color,
    )

    frame = resize(frame, new_height=edge_length)
    return frame


def resize(image, new_height):
    (current_height, current_width) = image.shape[:2]
    ratio = new_height / float(current_height)
    dim = (int(current_width * ratio), new_height)
    resized = cv2.resize(image, dim, interpolation=cv2.INTER_AREA)
    return resized


def apply_rotation(frame, rotation_count):
    if rotation_count == 0:
        pass
    elif rotation_count in [1, -3]:
        frame = cv2.rotate(frame, cv2.ROTATE_90_CLOCKWISE)
    elif rotation_count in [2, -2]:
        frame = cv2.rotate(frame, cv2.ROTATE_180)
    elif rotation_count in [-1, 3]:
        frame = cv2.rotate(frame, cv2.ROTATE_90_COUNTERCLOCKWISE)

    return frame


def prep_img_for_qpixmap(image: np.ndarray):
    """
    qpixmap needs dimensions divisible by 4 and without that weird things happen.
    """
    if image.shape[1] % 4 != 0:  # If the width of the row isn't divisible by 4
        padding_width = 4 - (image.shape[1] % 4)  # Calculate how much padding is needed
        padding = np.zeros(
            (image.shape[0], padding_width, image.shape[2]), dtype=image.dtype
        )  # Create a black image of the required size
        image = np.hstack([image, padding])  # Add the padding to the right of the image

    return image


def cv2_to_qimage(frame):
    image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    qt_frame = QImage(
        image.data,
        image.shape[1],
        image.shape[0],
        QImage.Format.Format_RGB888,
    )

    return qt_frame



# %%

from dataclasses import asdict
from datetime import datetime
from enum import Enum
from os.path import exists
from pathlib import Path

import cv2
import numpy as np
import rtoml

import caliscope.logger
from caliscope.calibration.capture_volume.capture_volume import CaptureVolume
from caliscope.calibration.capture_volume.point_estimates import PointEstimates
from caliscope.calibration.charuco import Charuco
from caliscope.cameras.camera import Camera
from caliscope.cameras.camera_array import CameraArray, CameraData

logger = caliscope.logger.get(__name__)


class ConfigSettings(Enum):
    """
    Control settings used to manage the processing of data
    """

    creation_date = "creation_date"
    camera_count = "camera_count"
    save_tracked_points_video = "save_tracked_points_video"
    fps_sync_stream_processing = "fps_sync_stream_processing"


# %%


class Configurator:
    """
    A factory to provide pre-configured objects and to save the configuration
    of currently existing objects.
    """

    def __init__(self, workspace_path: Path) -> None:
        self.workspace_path = workspace_path
        self.config_toml_path = Path(self.workspace_path, "config.toml")
        self.point_estimates_toml_path = Path(self.workspace_path, "point_estimates.toml")

        if exists(self.config_toml_path):
            self.refresh_config_from_toml()
            # this check only included for interfacing with historical tests...
            # if underlying tests data updated, this should be removed
            if "camera_count" not in self.dict.keys():
                self.dict["camera_count"] = 0
        else:
            logger.info("No existing config.toml found; creating starter file with charuco")
            self.dict = rtoml.loads("")
            self.dict[ConfigSettings.creation_date.value] = datetime.now()
            self.dict[ConfigSettings.camera_count.value] = 0
            self.dict[ConfigSettings.save_tracked_points_video.value] = True
            self.dict[ConfigSettings.fps_sync_stream_processing.value] = 100
            self.update_config_toml()

            # default values enforced below
            charuco = Charuco(4, 5, 11, 8.5, square_size_overide_cm=5.4)
            self.save_charuco(charuco)

    def save_camera_count(self, count):
        self.camera_count = count
        self.dict[ConfigSettings.camera_count.value] = count
        self.update_config_toml()

    def get_camera_count(self):
        return self.dict[ConfigSettings.camera_count.value]

    def get_save_tracked_points(self):
        if ConfigSettings.save_tracked_points_video.value not in self.dict.keys():
            return True
        else:
            return self.dict[ConfigSettings.save_tracked_points_video.value]

    def get_fps_sync_stream_processing(self):
        if ConfigSettings.fps_sync_stream_processing.value not in self.dict.keys():
            return 100
        else:
            return self.dict[ConfigSettings.fps_sync_stream_processing.value]

    def refresh_config_from_toml(self):
        logger.info("Populating config dictionary with config.toml data")
        # with open(self.config_toml_path, "r") as f:
        self.dict = rtoml.load(self.config_toml_path)

    def refresh_point_estimates_from_toml(self):
        logger.info("Populating config dictionary with point_estimates.toml data")
        # with open(self.config_toml_path, "r") as f:
        self.dict["point_estimates"] = rtoml.load(self.point_estimates_toml_path)

    def update_config_toml(self):
        # alphabetize by key to maintain standardized layout
        sorted_dict = {key: value for key, value in sorted(self.dict.items())}
        self.dict = sorted_dict

        dict_wo_point_estimates = {key: value for key, value in self.dict.items() if key != "point_estimates"}
        with open(self.config_toml_path, "w") as f:
            rtoml.dump(dict_wo_point_estimates, f)

    def save_capture_volume(self, capture_volume: CaptureVolume):
        # self.point_estimates = self.capture_volume.point_estimates
        # self.camera_array = self.capture_volume.camera_array
        self.save_camera_array(capture_volume.camera_array)
        self.save_point_estimates(capture_volume.point_estimates)

        self.dict["capture_volume"] = {}
        # self["capture_volume"]["RMSE_summary"] = self.capture_volume.rmse
        self.dict["capture_volume"]["stage"] = capture_volume.stage
        self.dict["capture_volume"]["origin_sync_index"] = capture_volume.origin_sync_index
        self.update_config_toml()

    def get_configured_camera_data(self) -> dict[CameraData]:
        all_camera_data = {}
        for key, params in self.dict.items():
            if key.startswith("cam_"):
                port = params["port"]

                if (
                    "error" in params.keys() and params["error"] is not None and params["error"] != "null"
                ):  # intrinsics have been calculated
                    error = params["error"]
                    matrix = np.array(params["matrix"])
                    distortions = np.array(params["distortions"])
                    grid_count = params["grid_count"]
                else:
                    error = None
                    matrix = None
                    distortions = None
                    grid_count = None

                if (
                    "translation" in params.keys()
                    and params["translation"] is not None
                    and params["translation"] != "null"
                ):  # Extrinsics have been calculated
                    translation = np.array(params["translation"])
                    rotation = np.array(params["rotation"])

                    if rotation.shape == (3,):  # camera rotation is stored as a matrix
                        rotation = cv2.Rodrigues(rotation)[0]

                else:
                    translation = None
                    rotation = None

                logger.info(f"Adding camera {port} to calibrated camera array...")
                cam_data = CameraData(
                    port=port,
                    size=params["size"],
                    rotation_count=params["rotation_count"],
                    error=error,
                    matrix=matrix,
                    distortions=distortions,
                    grid_count=grid_count,
                    translation=translation,
                    rotation=rotation,
                )

                all_camera_data[port] = cam_data
                logger.info(f"Camera successfully added at port {port}")
        logger.info("Camera data loaded and being passed back to caller")
        return all_camera_data

    def get_camera_array(self) -> CameraArray:
        """
        Load camera array directly from config file. The results of capture volume
        optimization and origin transformation will be reflected in this array
        which can then be the basis for future 3d point estimation
        """
        all_camera_data = self.get_configured_camera_data()
        camera_array = CameraArray(all_camera_data)
        logger.info("Camera array successfully created and being passed back to caller")
        return camera_array

    def get_point_estimates(self) -> PointEstimates:
        # only load point estimates into dictionary if saved more recently than last loaded

        if "point_estimates" not in self.dict.keys():
            self.refresh_point_estimates_from_toml()

        temp_data = self.dict["point_estimates"].copy()
        for key, value in temp_data.items():
            temp_data[key] = np.array(value)

        point_estimates = PointEstimates(**temp_data)

        return point_estimates

    def get_charuco(self) -> Charuco:
        """
        Charuco will always be available as it is created when initializing the config
        """

        logger.info("Loading charuco from config")
        params = self.dict["charuco"]

        # the below is intended to catch people using older configs without this info
        # if you come across this in mid 2025 or later, these may be safe to delete
        logger.info(f"charuco param are: {params}:")
        if "legacy_pattern" not in params.keys():
            params["legacy_pattern"]=False

        charuco = Charuco(
            columns=params["columns"],
            rows=params["rows"],
            board_height=params["board_height"],
            board_width=params["board_width"],
            dictionary=params["dictionary"],
            units=params["units"],
            aruco_scale=params["aruco_scale"],
            square_size_overide_cm=params["square_size_overide_cm"],
            inverted=params["inverted"],
            legacy_pattern=params["legacy_pattern"]
        )

        return charuco

    def save_charuco(self, charuco: Charuco):
        self.dict["charuco"] = charuco.__dict__
        logger.info(f"Saving charuco with params {charuco.__dict__} to config")
        self.update_config_toml()

    def save_camera(self, camera: Camera | CameraData):
        def none_or_list(value):
            # required to make sensible numeric format
            # otherwise toml formats as text
            if value is None or value == "null":
                return None
            else:
                return value.tolist()

        if camera.rotation is not None and camera.rotation != "null":
            # store rotation as 3 parameter rodrigues
            rotation_for_config = cv2.Rodrigues(camera.rotation)[0][:, 0]
            rotation_for_config = rotation_for_config.tolist()
        else:
            rotation_for_config = None

        params = {
            "port": camera.port,
            "size": camera.size,
            "rotation_count": camera.rotation_count,
            "error": camera.error,
            "matrix": none_or_list(camera.matrix),
            "distortions": none_or_list(camera.distortions),
            "translation": none_or_list(camera.translation),
            "rotation": rotation_for_config,
            "exposure": camera.exposure,
            "grid_count": camera.grid_count,
            # "ignore": camera.ignore,
            # "verified_resolutions": camera.verified_resolutions,
        }

        self.dict["cam_" + str(camera.port)] = params
        self.update_config_toml()

    def save_camera_array(self, camera_array: CameraArray):
        logger.info("Saving camera array....")
        for port, camera_data in camera_array.cameras.items():
            self.save_camera(camera_data)

    # Mac: leave this reference code in here for a potential splitting out of the recording functionality.
    # def get_cameras(self) -> dict[Camera]:
    #     cameras = {}

    #     def add_preconfigured_cam(params):
    #         # try:
    #         port = params["port"]
    #         logger.info(f"Attempting to add pre-configured camera at port {port}")

    #         if params["ignore"]:
    #             logger.info(f"Ignoring camera at port {port}")
    #             pass  # don't load it in
    #         else:
    #             if "verified_resolutions" in params.keys():
    #                 verified_resolutions = params["verified_resolutions"]
    #                 cameras[port] = Camera(port, verified_resolutions)
    #             else:
    #                 cameras[port] = Camera(port)

    #             camera = cameras[port]  # just for ease of reference
    #             camera.rotation_count = params["rotation_count"]
    #             camera.exposure = params["exposure"]

    #             if "error" in params.keys():  # then intrinsic params available
    #                 # if calibration done, then populate those as well
    #                 logger.info(
    #                     f"Adding in preconfigured intrinsic params at port {port}"
    #                 )
    #                 logger.info(f"Camera RMSE error for port {port}: {params['error']}")
    #                 camera.error = params["error"]
    #                 camera.matrix = np.array(params["matrix"]).astype(float)
    #                 camera.distortions = np.array(params["distortions"]).astype(float)
    #                 camera.grid_count = params["grid_count"]

    #             if "rotation" in params.keys():  # then extrinsic params available
    #                 logger.info(
    #                     f"Adding in preconfigured extrinsic params at port {port}"
    #                 )
    #                 camera.rotation = cv2.Rodrigues(np.array(params["rotation"]))[0]
    #                 camera.translation = np.array(params["translation"])

    #     with ThreadPoolExecutor() as executor:
    #         for key, params in self.dict.items():
    #             if key.startswith("cam"):
    #                 logger.info(f"Beginning to load {key} with params {params}")
    #                 executor.submit(add_preconfigured_cam, params)
    #     return cameras

    def save_point_estimates(self, point_estimates: PointEstimates):
        logger.info("Saving point estimates to toml...")

        temp_data = asdict(point_estimates)

        for key, params in temp_data.items():
            temp_data[key] = params.tolist()

        self.dict["point_estimates"] = temp_data

        with open(self.point_estimates_toml_path, "w") as f:
            rtoml.dump(self.dict["point_estimates"], f)


if __name__ == "__main__":
    import rtoml

    from caliscope import __app_dir__

    app_settings = rtoml.load(Path(__app_dir__, "settings.toml"))
    recent_projects: list = app_settings["recent_projects"]

    recent_project_count = len(recent_projects)
    session_path = Path(recent_projects[recent_project_count - 1])

    config = Configurator(session_path)

# %%



import cv2
import numpy as np


def apply_rotation(frame: np.ndarray, rotation_count: int) -> np.ndarray:
    if rotation_count == 0:
        pass
    elif rotation_count in [1, -3]:
        frame = cv2.rotate(frame, cv2.ROTATE_90_CLOCKWISE)
    elif rotation_count in [2, -2]:
        frame = cv2.rotate(frame, cv2.ROTATE_180)
    elif rotation_count in [-1, 3]:
        frame = cv2.rotate(frame, cv2.ROTATE_90_COUNTERCLOCKWISE)

    return frame


def unrotate_points(xy: np.ndarray, rotation_count: int, frame_width: int, frame_height: int) -> np.ndarray:
    xy_unrotated = xy.copy()

    if rotation_count == 0 or len(xy) == 0:
        pass
    elif rotation_count in [1, -3]:
        # Reverse of 90 degrees clockwise rotation
        xy_unrotated[:, 0], xy_unrotated[:, 1] = xy[:, 1], frame_width - xy[:, 0]

    elif rotation_count in [2, -2]:
        # NOTE: have not verified this with a test case
        # Reverse of 180 degrees rotation
        xy_unrotated[:, 0], xy_unrotated[:, 1] = (
            frame_width - xy[:, 0],
            frame_height - xy[:, 1],
        )
    elif rotation_count in [-1, 3]:
        # Reverse of 90 degrees counter-clockwise rotation
        xy_unrotated[:, 0], xy_unrotated[:, 1] = frame_height - xy[:, 1], xy[:, 0]

    return xy_unrotated



import math
from time import sleep

from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import (
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QScrollArea,
    QVBoxLayout,
    QWidget,
)

import caliscope.logger
from caliscope.gui.frame_emitters.frame_dictionary_emitter import FrameDictionaryEmitter
from caliscope.synchronized_stream_manager import SynchronizedStreamManager

logger = caliscope.logger.get(__name__)


class SynchedFramesDisplay(QWidget):
    """
    This widget is not intended to have any interactive functionality at all and to only
    provide a window to the user of the current landmark tracking

    This is why the primary input is the sync stream manager directly and not the controller
    Apologies to Future Mac who is reading this and regretting my decisions.
    """

    def __init__(self, sync_stream_manager: SynchronizedStreamManager):
        super(SynchedFramesDisplay, self).__init__()

        self.setWindowTitle("Tracking Landmarks....")
        self.sync_stream_manager = sync_stream_manager
        self.synchronizer = self.sync_stream_manager.synchronizer
        self.ports = self.synchronizer.ports

        while not hasattr(sync_stream_manager.synchronizer, "current_sync_packet"):
            logger.info("Waiting for synchronizer to have sync packet")
            sleep(0.25)

        # create tools to build and emit the displayed frame
        self.frame_dictionary_emitter = FrameDictionaryEmitter(
            self.synchronizer, self.sync_stream_manager.all_camera_data
        )
        self.frame_dictionary_emitter.start()

        # all video output routed to qlabels stored in a dictionariy
        # make it as square as you can get it
        self.recording_displays = {str(port): QLabel() for port in self.ports}

        self.place_widgets()
        self.connect_widgets()

        logger.info("Synched Frames Display init complete")

    def place_widgets(self):
        self.scroll_area = QScrollArea()
        grid_widget = QWidget()
        grid_widget.setLayout(QGridLayout())

        camera_count = len(self.ports)
        grid_columns = int(math.ceil(camera_count**0.5))
        row = 0
        column = 0
        for port in sorted(self.ports):
            grid_widget.layout().addWidget(self.recording_displays[str(port)], row, column)

            # update row and column for next iteration
            if column >= grid_columns - 1:
                # start fresh on next row
                column = 0
                row += 1
            else:
                column += 1

        grid_widget.setLayout(grid_widget.layout())
        padded_grid_layout = QHBoxLayout()
        padded_grid_layout.addStretch(1)
        padded_grid_layout.addWidget(grid_widget)
        padded_grid_layout.addStretch(1)

        scroll_viewport = QWidget()
        scroll_viewport.setLayout(padded_grid_layout)

        # Scroll Area Properties
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setWidget(scroll_viewport)
        # self.setCentralWidget(self.scroll_area)
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.scroll_area)

    def connect_widgets(self):
        self.frame_dictionary_emitter.FramesBroadcast.connect(self.ImageUpdateSlot)
        # self.frame_dictionary_emitter.dropped_fps.connect(self.update_dropped_fps)
        self.frame_dictionary_emitter.close_window.connect(self.close)

    # @Slot(dict)
    # def update_dropped_fps(self, dropped_fps: dict):
    #     "Unravel dropped fps dictionary to a more readable string"
    #     text = "Rate of Frame Dropping by Port:    "
    #     for port, drop_rate in dropped_fps.items():
    #         text += f"{port}: {drop_rate:.0%}        "
    #     self.dropped_fps_label.setText(text)

    @Slot(dict)
    def ImageUpdateSlot(self, qpixmaps: dict):
        # logger.info(f"About to process q_image_dict: {q_image_dict}")
        for port, qpixmap in qpixmaps.items():
            # qpixmap = QPixmap.fromImage(qpixmap)
            logger.debug("About to set qpixmap to display")
            self.recording_displays[str(port)].setPixmap(qpixmap)
            logger.debug("successfully set display")



import logging
from pathlib import Path
from queue import Queue
from threading import Event, Thread
from time import perf_counter, sleep

import cv2
import numpy as np
import pandas as pd

import caliscope.logger
from caliscope.packets import FramePacket, Tracker

logger = caliscope.logger.get(__name__)
logger.setLevel(logging.INFO)


class RecordedStream:
    """
    Analogous to the live stream, this will place frames on a queue
    These can then be harvested and synchronized by a Synchronizer
    Within the stream, point detection occurs.
    """

    def __init__(
        self,
        directory: Path,
        port: int,
        # size: tuple = None,
        rotation_count: int = 0,
        fps_target: int = None,
        tracker: Tracker = None,
        break_on_last=True,
    ):
        # self.port = port
        self.directory = directory
        self.port = port
        self.rotation_count = rotation_count

        # stop while loop if end reached.
        # Preferred behavior for automated file processing, not interactive frame selection
        self.break_on_last = break_on_last

        self.tracker = tracker

        video_path = str(Path(self.directory, f"port_{self.port}.mp4"))
        self.capture = cv2.VideoCapture(video_path)

        # for playback, set the fps target to the actual
        self.original_fps = int(self.capture.get(cv2.CAP_PROP_FPS))
        if fps_target is None:
            fps_target = self.original_fps

        width = int(self.capture.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(self.capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.size = (width, height)

        self.stop_event = Event()
        self._jump_q = Queue(maxsize=1)
        self._pause_event = Event()
        self._pause_event.clear()
        self.subscribers = []

        ############ PROCESS WITH TRUE TIME STAMPS IF AVAILABLE #########################
        synched_frames_history_path = Path(self.directory, "frame_time_history.csv")

        if synched_frames_history_path.exists():
            synched_frames_history = pd.read_csv(synched_frames_history_path)

            self.port_history = synched_frames_history[synched_frames_history["port"] == self.port]

            self.port_history["frame_index"] = self.port_history["frame_time"].rank(method="min").astype(int) - 1

        ########### INFER TIME STAMP IF NOT AVAILABLE ####################################
        else:
            frame_count = int(self.capture.get(cv2.CAP_PROP_FRAME_COUNT))
            mocked_port_history = {
                "frame_index": [i for i in range(0, frame_count)],
                "frame_time": [i / self.original_fps for i in range(0, frame_count)],
            }
            self.port_history = pd.DataFrame(mocked_port_history)

        # note that this is not simply 0 and frame count because the syncronized recording might start recording many
        # frames into pulling from a camera
        # this is one of those unhappy artifacts that may be a good candidate for simplification in a future refactor
        self.start_frame_index = self.port_history["frame_index"].min()
        self.last_frame_index = self.port_history["frame_index"].max()

        # initialize properties
        self.frame_index = 0
        self.frame_time = 0
        self.set_fps_target(fps_target)

    # def set_tracking_on(self, track: bool):
    #     if track:
    #         logger.info(f"Turning tracking on for recorded stream {self.port}")
    #         self.track_points =
    #     else:
    #         logger.info(f"Turning tracking off for recorded stream {self.port}")
    #         self.track_points.clear()

    def subscribe(self, queue: Queue):
        if queue not in self.subscribers:
            logger.info(f"Adding queue to subscribers at recorded stream {self.port}")
            self.subscribers.append(queue)
            logger.info(f"...now {len(self.subscribers)} subscriber(s) at {self.port}")
        else:
            logger.warning(f"Attempted to subscribe to recorded stream at port {self.port} twice")

    def unsubscribe(self, queue: Queue):
        if queue in self.subscribers:
            logger.info(f"Removing subscriber from queue at recorded stream {self.port}")
            self.subscribers.remove(queue)
            logger.info(f"{len(self.subscribers)} subscriber(s) remain at recorded stream {self.port}")
        else:
            logger.warning(
                f"Attempted to unsubscribe to recorded stream that was not subscribed to\
                at port {self.port} twice"
            )

    def set_fps_target(self, fps):
        self.fps = fps
        if self.fps is None:
            self.milestones = None
        else:
            milestones = []
            for i in range(0, fps):
                milestones.append(i / fps)
            logger.info(f"Setting fps to {self.fps}")
            self.milestones = np.array(milestones)

    def wait_to_next_frame(self):
        """
        based on the next milestone time, return the time needed to sleep so that
        a frame read immediately after would occur when needed
        """

        time = perf_counter()
        fractional_time = time % 1
        all_wait_times = self.milestones - fractional_time
        future_wait_times = all_wait_times[all_wait_times > 0]

        if len(future_wait_times) == 0:
            return 1 - fractional_time
        else:
            return future_wait_times[0]

    def jump_to(self, frame_index: int):
        logger.info(f"Placing {frame_index} on jump q to reset capture position")
        self._jump_q.put(frame_index)

    def pause(self):
        logger.info(f"Pausing recorded stream at port {self.port}")
        self._pause_event.set()

    def unpause(self):
        logger.info(f"Unpausing recorded stream at port {self.port}")
        self._pause_event.clear()

    def play_video(self):
        logger.info(f"Initiating _play_worker for Camera {self.port}")
        self.thread = Thread(target=self._play_worker, args=[], daemon=False)
        self.thread.start()

    def _play_worker(self):
        """
        Places FramePacket on the out_q, mimicking the behaviour of the LiveStream.
        """

        self.frame_index = self.start_frame_index
        logger.info(f"Beginning playback of video for port {self.port}")

        while not self.stop_event.is_set():
            current_frame = self.port_history["frame_index"] == self.frame_index
            self.frame_time = self.port_history[current_frame]["frame_time"]
            self.frame_time = float(self.frame_time)

            ########## BEGIN NO SUBSCRIBERS SPINLOCK ##################
            spinlock_looped = False
            while len(self.subscribers) == 0 and not self.stop_event.is_set():
                if not spinlock_looped:
                    logger.info(f"Spinlock initiated at port {self.port}")
                    spinlock_looped = True
                sleep(0.5)
            if spinlock_looped:
                logger.info(f"Spinlock released at port {self.port}")
            ########## END NO SUBSCRIBERS SPINLOCK ##################

            if self.milestones is not None:
                sleep(self.wait_to_next_frame())
            # logger.info(f"about to read frame {self.frame_index} from capture at port {self.port}")
            success, self.frame = self.capture.read()

            if not success:
                break

            if self.tracker is not None:
                self.point_data = self.tracker.get_points(self.frame, self.port, self.rotation_count)
                draw_instructions = self.tracker.scatter_draw_instructions
            else:
                self.point_data = None
                draw_instructions = None

            frame_packet = FramePacket(
                port=self.port,
                frame_index=self.frame_index,
                frame_time=self.frame_time,
                frame=self.frame,
                points=self.point_data,
                draw_instructions=draw_instructions,
            )

            logger.debug(
                f"Placing frame on q {self.port} for frame time: {self.frame_time} and frame index: {self.frame_index}"
            )

            for q in self.subscribers:
                q.put(frame_packet)

            logger.debug(f"Incrementing frame index from {self.frame_index} to {self.frame_index+1}")
            self.frame_index += 1

            if self.frame_index > self.last_frame_index and self.break_on_last:
                logger.info(f"Ending recorded playback at port {self.port}")
                # time of -1 indicates end of stream
                frame_packet = FramePacket(
                    port=self.port,
                    frame_index=-1,
                    frame_time=-1,
                    frame=None,
                    points=None,
                )

                for q in self.subscribers:
                    q.put(frame_packet)
                break

            ############ Autopause if last frame and in playback mode (i.e. break_on_last == False)
            if not self.break_on_last and self.frame_index > self.last_frame_index:
                self.frame_index = self.last_frame_index
                self._pause_event.set()

            ############ SPIN LOCK FOR PAUSE ##################
            pause_logged = False
            while self._pause_event.is_set():
                # logger.info("I'm paused")
                if not pause_logged:
                    logger.info("Initiating Pause")
                    pause_logged = True

                if not self._jump_q.empty():
                    logger.info("New Value on jump queue, exiting pause spin lock")
                    break

                sleep(0.1)
            #######################################################
            if not self._jump_q.empty():
                self.frame_index = self._jump_q.get()
                logger.info(f"Setting port {self.port} capture object to frame index {self.frame_index}")
                self.capture.set(cv2.CAP_PROP_POS_FRAMES, self.frame_index)



# %%

from collections import OrderedDict
from dataclasses import dataclass

import cv2
import numpy as np
from numba.typed import Dict

import caliscope.logger

logger = caliscope.logger.get(__name__)
CAMERA_PARAM_COUNT = 6


@dataclass
class CameraData:
    """
    A place to hold the calibration data associated with a camera that has been populated from a config file.
    For use with final setting of the array and triangulation, but no actual camera management.
    Not loving the way this is implemented as an adjunct to the Camera object, but here we are
    """

    port: int
    size: tuple
    rotation_count: int = 0
    error: float = None  # the RMSE of reprojection associated with the intrinsic calibration
    matrix: np.ndarray = None
    distortions: np.ndarray = None  #
    exposure: int = None
    grid_count: int = None
    ignore: bool = False
    verified_resolutions: np.ndarray = None
    translation: np.ndarray = None  # camera relative to world
    rotation: np.ndarray = None  # camera relative to world

    @property
    def transformation(self):
        """ "
        Rotation and transformation combined to allow
        """

        t = np.hstack([self.rotation, np.expand_dims(self.translation, 1)])
        t = np.vstack([t, np.array([0, 0, 0, 1], np.float32)])
        return t

    @transformation.setter
    def transformation(self, t: np.ndarray):
        self.rotation = t[0:3, 0:3]
        self.translation = t[0:3, 3]
        logger.info(f"Rotation and Translation being updated to {self.rotation} and {self.translation}")

    @property
    def projection_matrix(self):
        return self.matrix @ self.transformation[0:3, :]

    def extrinsics_to_vector(self):
        """
        Converts camera parameters to a numpy vector for use with bundle adjustment.
        """
        # rotation of the camera relative to the world
        rotation_rodrigues = cv2.Rodrigues(self.rotation)[0]  # elements 0,1,2
        port_param = np.hstack([rotation_rodrigues[:, 0], self.translation])

        return port_param

    def extrinsics_from_vector(self, row):
        """
        Takes a vector in the same format that is output of `extrinsics_to_vector` and updates the camera
        """

        # convert back to world frame of reference
        self.rotation = cv2.Rodrigues(row[0:3])[0]
        self.translation = np.array([row[3:6]], dtype=np.float64)[0]

    def get_display_data(self) -> OrderedDict:
        # Extracting camera matrix parameters
        # self.matrix = None
        if self.matrix is not None:
            fx, fy = self.matrix[0, 0], self.matrix[1, 1]
            cx, cy = self.matrix[0, 2], self.matrix[1, 2]
        else:
            fx, fy = None, None
            cx, cy = None, None

        # Extracting distortion coefficients
        if self.distortions is not None:
            k1, k2, p1, p2, k3 = self.distortions.ravel()[:5]
        else:
            k1, k2, p1, p2, k3 = None, None, None, None, None

        def round_or_none(value, places):
            if value is None:
                return None
            else:
                return round(value, places)

        # Creating the dictionary with OrderedDict
        camera_display_dict = OrderedDict(
            [
                ("size", self.size),
                ("RMSE", self.error),
                ("Grid_Count", self.grid_count),
                ("rotation_count", self.rotation_count),
                (
                    "intrinsic_parameters",
                    OrderedDict(
                        [
                            ("focal_length_x", round_or_none(fx, 2)),
                            ("focal_length_y", round_or_none(fy, 2)),
                            ("optical_center_x", round_or_none(cx, 2)),
                            ("optical_center_y", round_or_none(cy, 2)),
                        ]
                    ),
                ),
                (
                    "distortion_coefficients",
                    OrderedDict(
                        [
                            ("radial_k1", round_or_none(k1, 2)),
                            ("radial_k2", round_or_none(k2, 2)),
                            ("radial_k3", round_or_none(k3, 2)),
                            ("tangential_p1", round_or_none(p1, 2)),
                            ("tangential_p2", round_or_none(p2, 2)),
                        ]
                    ),
                ),
            ]
        )

        return camera_display_dict

    def erase_calibration_data(self):
        self.error = None
        self.matrix = None
        self.distortions = None
        self.grid_count = None
        self.translation = None
        self.rotation = None


@dataclass
class CameraArray:
    """The plan is that this will expand to become an interface for setting the origin.
    At the moment all it is doing is holding a dictionary of CameraData objects"""

    cameras: dict

    @property
    def port_index(self):
        """
        Provides a dictionary mapping the camera port to an index. Generally,this
        will match the camera ports 1:1, but will be different when a camera
        is being ignored. Used to manage reference to camera parameters in xy_reprojection_error
        used within the least_squares optimization of the capture volume.
        """
        not_ignored_ports = [port for port, cam in self.cameras.items() if not cam.ignore]
        not_ignored_ports.sort()
        not_ignored_indices = [i for i in range(len(not_ignored_ports))]
        port_indices = {port: i for port, i in zip(not_ignored_ports, not_ignored_indices)}

        return port_indices

    @property
    def index_port(self):
        return {value: key for key, value in self.port_index.items()}

    def get_extrinsic_params(self):
        """for each camera build the CAMERA_PARAM_COUNT element parameter index
        camera_params with shape (n_cameras, CAMERA_PARAM_COUNT)
        contains initial estimates of parameters for all cameras.
        First 3 components in each row form a rotation vector (https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula),
        next 3 components form a translation vector
        """

        camera_params = None
        # ensure that parameters are built up in order of the corresponding index
        for index in sorted(self.index_port.keys()):
            port = self.index_port[index]
            cam = self.cameras[port]
            port_param = cam.extrinsics_to_vector()

            if camera_params is None:
                camera_params = port_param
            else:
                camera_params = np.vstack([camera_params, port_param])

        return camera_params

    def update_extrinsic_params(self, least_sq_result_x: np.array):
        n_cameras = len(self.port_index)
        n_cam_param = 6  # 6 DoF
        flat_camera_params = least_sq_result_x[0 : n_cameras * n_cam_param]
        new_camera_params = flat_camera_params.reshape(n_cameras, n_cam_param)

        # update camera array with new positional data
        for index in range(len(new_camera_params)):
            port = self.index_port[index]  # correct in case ignoring a camera
            cam_vec = new_camera_params[index, :]
            self.cameras[port].extrinsics_from_vector(cam_vec)

    def all_extrinsics_calibrated(self) -> bool:
        # assume extrinsics calibrated and provide otherwise
        full_extrinsics = True
        for port, cam in self.cameras.items():
            if cam.rotation is None or cam.translation is None:
                full_extrinsics = False
        return full_extrinsics

    def all_intrinsics_calibrated(self) -> bool:
        # assume true and prove false
        full_intrinsics = True
        for port, cam in self.cameras.items():
            if cam.matrix is None or cam.distortions is None:
                full_intrinsics = False
        return full_intrinsics

    @property
    def projection_matrices(self) -> Dict:
        logger.info("Creating camera array projection matrices")
        proj_mat = Dict()
        for port, cam in self.cameras.items():
            proj_mat[port] = cam.projection_matrix

        return proj_mat



import shutil
from pathlib import Path

import caliscope.logger

logger = caliscope.logger.get(__name__)


def copy_contents(src_folder, dst_folder):
    """
    Helper function to port a test case data folder over to a temp directory
    used for testing purposes so that the test case data doesn't get overwritten
    """
    src_path = Path(src_folder)
    dst_path = Path(dst_folder)

    if dst_path.exists():
        shutil.rmtree(dst_path)

    # Create the destination folder if it doesn't exist
    dst_path.mkdir(parents=True, exist_ok=False)

    for item in src_path.iterdir():
        # Construct the source and destination paths
        src_item = src_path / item
        dst_item = dst_path / item.name

        # Copy file or directory
        if src_item.is_file():
            logger.info(f"Copying file at {src_item} to {dst_item}")
            shutil.copy2(src_item, dst_item)  # Copy file preserving metadata

        elif src_item.is_dir():
            logger.info(f"Copying directory at {src_item} to {dst_item}")
            shutil.copytree(src_item, dst_item)



from pathlib import Path

import numpy as np
import pandas as pd

import caliscope.logger
from caliscope.calibration.capture_volume.helper_functions.get_stereotriangulated_table import (
    get_stereotriangulated_table,
)
from caliscope.calibration.capture_volume.point_estimates import PointEstimates
from caliscope.cameras.camera_array import CameraArray

logger = caliscope.logger.get(__name__)


def get_points_2d_df(stereotriangulated_table: pd.DataFrame):
    points_2d_port_A = stereotriangulated_table[["port_A", "sync_index", "point_id", "x_A", "y_A"]].rename(
        columns={
            "port_A": "camera",
            "sync_index": "sync_index",
            "point_id": "corner_id",
            "x_A": "x_2d",
            "y_A": "y_2d",
        }
    )

    points_2d_port_B = stereotriangulated_table[["port_B", "sync_index", "point_id", "x_B", "y_B"]].rename(
        columns={
            "port_B": "camera",
            "sync_index": "sync_index",
            "point_id": "corner_id",
            "x_B": "x_2d",
            "y_B": "y_2d",
        }
    )

    points_2d_df = (
        pd.concat([points_2d_port_A, points_2d_port_B])
        .drop_duplicates()
        .sort_values(["sync_index", "corner_id", "camera"])
        .rename(columns={"sync_index": "sync_index"})
    )
    return points_2d_df


# get 3d points with indices to merge back into points_2d
def get_points_3d_df(stereotriangulated_table):
    points_3d_df = (
        stereotriangulated_table[["sync_index", "point_id", "pair", "x_pos", "y_pos", "z_pos"]]
        .sort_values(["sync_index", "point_id"])
        .groupby(["sync_index", "point_id"])
        .agg({"x_pos": "mean", "y_pos": "mean", "z_pos": "mean", "pair": "size"})
        .rename(columns={"pair": "count", "x_pos": "x_3d", "y_pos": "y_3d", "z_pos": "z_3d"})
        .reset_index()
        .reset_index()
        .rename(columns={"index": "index_3d", "point_id": "corner_id"})
    )
    return points_3d_df


def get_merged_2d_3d(stereotriangulated_table):
    """
    For each 2d point line, add in the estimated 3d point position
    """
    points_2d_df = get_points_2d_df(stereotriangulated_table)
    points_3d_df = get_points_3d_df(stereotriangulated_table)

    merged_point_data = (
        points_2d_df.merge(points_3d_df, how="left", on=["sync_index", "corner_id"])
        .sort_values(["camera", "sync_index", "corner_id"])
        .dropna()
    )

    return merged_point_data


def get_point_estimates(camera_array: CameraArray, point_data_path: Path) -> PointEstimates:
    """
    Stereotriangulates data from xy_charuco to generate initial x,y,z estimates
    formats the data into a PointEstimates object that has the
    data structured in a way that is amenable to bundle adjustment
    """

    logger.info("Creating point history object based on camera_array and stereotriangulated_table")
    stereotriangulated_points = get_stereotriangulated_table(camera_array, point_data_path)

    points_3d_df = get_points_3d_df(stereotriangulated_points)
    merged_point_data = get_merged_2d_3d(stereotriangulated_points)

    camera_indices = np.array(merged_point_data["camera"], dtype=np.int64)
    img = np.array(merged_point_data[["x_2d", "y_2d"]])
    corner_id = np.array(merged_point_data["corner_id"], dtype=np.int64)
    obj_indices = np.array(merged_point_data["index_3d"], dtype=np.int64)
    sync_index = np.array(merged_point_data["sync_index"], dtype=np.int64)
    obj = np.array(points_3d_df[["x_3d", "y_3d", "z_3d"]])
    # obj_corner_id = np.array(points_3d_df[["corner_id"]])

    return PointEstimates(
        sync_indices=sync_index,
        camera_indices=camera_indices,
        point_id=corner_id,
        img=img,
        obj_indices=obj_indices,
        obj=obj,
        # obj_corner_id=obj_corner_id,
    )



import sys

import caliscope.logger
from caliscope.gui.main_widget import launch_main

logger = caliscope.logger.get(__name__)


def CLI_parser():
    if len(sys.argv) == 1:
        launch_main()

    if len(sys.argv) == 2:
        launch_widget = sys.argv[1]

        # if launch_widget in ["calibrate", "cal", "-c"]:
        #     launch_extrinsic_calibration_widget(session_path)

        if launch_widget in ["record", "rec", "-r"]:
            pass
            # launch_recording_widget(session_path)



# from PySide6.QtCore import QObject, Signal
from pathlib import Path
from queue import Queue
from threading import Event, Thread

import cv2
import pandas as pd

import caliscope.logger
from caliscope.cameras.synchronizer import Synchronizer
from caliscope.packets import SyncPacket

logger = caliscope.logger.get(__name__)


class VideoRecorder:
    def __init__(self, synchronizer: Synchronizer, suffix: str = None):
        """
        suffix: provide a way to clarify any modifications to the video that are being saved
        This is likely going to be the name of the tracker used in most cases
        """
        super().__init__()
        self.synchronizer = synchronizer

        # set text to be appended as port_X_{suffix}.mp4
        # will also be appended to xy_{suffix}
        if suffix is not None:
            self.suffix = "_" + suffix
        else:
            self.suffix = ""

        self.recording = False
        self.sync_index = 0  # no sync packets at init... absence of initialized value can cause errors elsewhere
        # build dict that will be stored to csv
        self.trigger_stop = Event()

        self.sync_packet_in_q = Queue(-1)

    def build_video_writers(self):
        """
        suffix provides a way to provide additional labels to the mp4 file name
        This would be relevant when performing post-processing and saving out frames with points
        """
        # create a dictionary of videowriters
        self.video_writers = {}
        for port, stream in self.synchronizer.streams.items():
            path = str(Path(self.destination_folder, f"port_{port}{self.suffix}.mp4"))
            logger.info(f"Building video writer for port {port}; recording to {path}")
            fourcc = cv2.VideoWriter_fourcc(*"MP4V")
            frame_size = stream.size
            logger.info(f"Creating video writer with fps of {stream.original_fps} and frame size of {frame_size}")
            writer = cv2.VideoWriter(path, fourcc, stream.original_fps, frame_size)
            self.video_writers[port] = writer

    def save_data_worker(self, include_video: bool, show_points: bool, store_point_history: bool):
        # connect video recorder to synchronizer via an "in" queue
        if include_video:
            self.build_video_writers()

        # I think I put this here so that it will get reset if you reuse the same recorder..
        self.frame_history = {
            "sync_index": [],
            "port": [],
            "frame_index": [],
            "frame_time": [],
        }

        self.point_data_history = {
            "sync_index": [],
            "port": [],
            "frame_index": [],
            "frame_time": [],
            "point_id": [],
            "img_loc_x": [],
            "img_loc_y": [],
            "obj_loc_x": [],
            "obj_loc_y": [],
        }

        self.synchronizer.subscribe_to_sync_packets(self.sync_packet_in_q)
        syncronizer_subscription_released = False

        # this is where the issue is... need to figure out when the queue is empty...
        logger.info("Entering Save data worker loop entered")
        while self.sync_packet_in_q.qsize() > 0 or not self.trigger_stop.is_set():
            sync_packet: SyncPacket = self.sync_packet_in_q.get()

            # provide periodic updates of recording queue
            logger.debug("Getting size of sync packet q")
            backlog = self.sync_packet_in_q.qsize()
            if backlog % 25 == 0 and backlog != 0:
                logger.info(f"Size of unsaved frames on the recording queue is {self.sync_packet_in_q.qsize()}")

            if sync_packet is None:
                # relenvant when
                logger.info("End of sync packets signaled...breaking record loop")
                break

            self.sync_index = sync_packet.sync_index

            for port, frame_packet in sync_packet.frame_packets.items():
                if frame_packet is not None:
                    logger.debug("Processiong frame packet...")
                    # read in the data for this frame for this port
                    if show_points:
                        frame = frame_packet.frame_with_points
                    else:
                        frame = frame_packet.frame

                    frame_index = frame_packet.frame_index
                    frame_time = frame_packet.frame_time

                    if include_video:
                        # store the frame
                        if self.sync_index % 50 == 0:
                            logger.debug(f"Writing frame for port {port} and sync index {self.sync_index}")
                            logger.debug(f"frame size  {frame.shape}")

                        self.video_writers[port].write(frame)

                        # store to assocated data in the dictionary
                        self.frame_history["sync_index"].append(self.sync_index)
                        self.frame_history["port"].append(port)
                        self.frame_history["frame_index"].append(frame_index)
                        self.frame_history["frame_time"].append(frame_time)

                    new_tidy_table = frame_packet.to_tidy_table(self.sync_index)
                    if new_tidy_table is not None:  # i.e. it has data
                        for key, value in self.point_data_history.copy().items():
                            logger.debug("Extending tidy table of point history")
                            self.point_data_history[key].extend(new_tidy_table[key])

            if not syncronizer_subscription_released and self.trigger_stop.is_set():
                logger.info("Save frame worker winding down...")
                syncronizer_subscription_released = True
                self.synchronizer.release_sync_packet_q(self.sync_packet_in_q)
                # self.sync_packet_in_q = Queue(-1)
                # self.recording_stop_signal.emit()

        # a proper release is strictly necessary to ensure file is readable
        if include_video:
            logger.info("releasing video writers...")
            for port in self.synchronizer.ports:
                logger.info(f"releasing video writer for port {port}")
                self.video_writers[port].release()

            # del self.video_writers

            logger.info("Initiate storing of frame history")
            self.store_frame_history()

        logger.info("Initiate storing of point history")
        if store_point_history:
            self.store_point_history()
        self.trigger_stop.clear()  # reset stop recording trigger
        self.recording = False
        logger.info("About to emit `all frames saved` signal")
        # self.all_frames_saved_signal.emit()

    def store_point_history(self):
        df = pd.DataFrame(self.point_data_history)
        point_data_path = str(Path(self.destination_folder, f"xy{self.suffix}.csv"))
        logger.info(f"Storing point data in {point_data_path}")
        df.to_csv(point_data_path, index=False, header=True)

    def store_frame_history(self):
        df = pd.DataFrame(self.frame_history)
        frame_hist_path = str(Path(self.destination_folder, "frame_time_history.csv"))
        logger.info(f"Storing frame history to {frame_hist_path}")
        df.to_csv(frame_hist_path, index=False, header=True)

    def store_active_config(self):
        pass

    def start_recording(
        self,
        destination_folder: Path,
        include_video=True,
        show_points=False,
        store_point_history=True,
    ):
        """
        Option exists to not store video if only interested in getting points from original video

        Parent of destination folder will be the source of the config file that will be stored with the video
        This enables the nested processing of videos (i.e. Recording_1 will store the main config.toml,
        then POSE subfolder will store config.toml from Recording_1). Each folder should largely become self
        contained and portable for analysis / reconstruction.
        """
        logger.info(f"All video data to be saved to {destination_folder}")

        self.destination_folder = destination_folder
        # create the folder if it doesn't already exist
        self.destination_folder.mkdir(exist_ok=True, parents=True)
        self.recording = True
        self.recording_thread = Thread(
            target=self.save_data_worker,
            args=[include_video, show_points, store_point_history],
            daemon=True,
        )
        self.recording_thread.start()

    def stop_recording(self):
        logger.info("about to Stop recording initiated within VideoRecorder")
        self.trigger_stop.set()
        logger.info("Stop recording initiated within VideoRecorder")


def find_config_file(start_dir):
    """
    Search for a 'config.toml' file starting from 'start_dir' and moving up to the parent directories.

    :param start_dir: Pathlib Path object of the starting directory
    :return: Path object of the found 'config.toml' file or None if not found
    """
    current_dir = start_dir

    while True:
        config_file = current_dir / "config.toml"
        if config_file.is_file():
            return config_file
        if current_dir.parent == current_dir:
            # We have reached the root directory without finding the file
            break
        current_dir = current_dir.parent

    return None



import csv
from pathlib import Path

import pandas as pd

import caliscope.logger
from caliscope.packets import Tracker

logger = caliscope.logger.get(__name__)


def xyz_to_wide_labelled(xyz: pd.DataFrame, tracker: Tracker) -> pd.DataFrame:
    """
    Will save a csv file in the same directory as the long xyz point data
    Column headings will be based on the point_id names in the Tracker
    """

    # save out named data in a tabular format
    xyz = xyz.rename(
        {
            "x_coord": "x",
            "y_coord": "y",
            "z_coord": "z",
        },
        axis=1,
    )
    xyz = xyz[["sync_index", "point_id", "x", "y", "z"]]

    xyz["point_name"] = xyz["point_id"].map(tracker.get_point_name)
    # pivot the DataFrame wider
    df_wide = xyz.pivot_table(index=["sync_index"], columns="point_name", values=["x", "y", "z"])
    # flatten the column names
    df_wide.columns = ["{}_{}".format(y, x) for x, y in df_wide.columns]
    # reset the index
    df_wide = df_wide.reset_index()
    # merge the rows with the same sync_index
    df_merged = df_wide.groupby("sync_index").agg("first")
    # sort the dataframe
    df_merged = df_merged.sort_index(axis=1, ascending=True)
    return df_merged


def xyz_to_trc(xyz: pd.DataFrame, tracker: Tracker, time_history_path: Path, target_path: Path):
    """
    Will save a .trc file in the same folder as the long xyz data
    relies on xyz_to_wide_csv for input data
    """
    # create xyz_labelled file to provide input for trc creation
    xyz_labelled = xyz_to_wide_labelled(xyz, tracker)

    # from here I need to get a .trc file format. For part of that I also need to know the framerate.
    # time_history_path = Path(target_path.parent, "frame_time_history.csv")
    time_history = pd.read_csv(time_history_path)

    # get the mean time by sync index
    # Group by 'sync_index' and calculate mean 'frame_time'
    sync_time = time_history.groupby("sync_index")["frame_time"].mean()

    # Shift times so that it starts at zero
    min_time = sync_time.min()
    sync_time = round(sync_time - min_time, 3)
    xyz_labelled.insert(1, "mean_frame_time", sync_time)
    # %%

    # Calculate time differences between consecutive frames
    xyz_labelled.sort_values(by="mean_frame_time", inplace=True)
    xyz_labelled["time_diff"] = xyz_labelled["mean_frame_time"].diff()

    # Calculate frame rate for each pair of frames (avoid division by zero)
    xyz_labelled["frame_rate"] = xyz_labelled["time_diff"].apply(lambda x: 1 / x if x != 0 else 0)

    # Calculate mean frame rate (drop the first value which is NaN due to the diff operation)
    mean_frame_rate = xyz_labelled["frame_rate"].dropna().mean()

    # Rename 'sync_index' to 'Frame' and 'mean_frame_time' to 'Time'
    xyz_labelled = xyz_labelled.reset_index()  # need sync_index to be just a regular column
    xyz_labelled.rename(columns={"sync_index": "Frame", "mean_frame_time": "Time"}, inplace=True)
    xyz_labelled.drop(columns=["time_diff", "frame_rate"], inplace=True)  # no longer needed

    # Make sure all following fields are in alphabetical order
    # First, get the columns to be sorted, i.e., all columns excluding 'Frame' and 'Time'
    cols_to_sort = xyz_labelled.columns.tolist()[2:]
    cols_to_sort = [col for col in cols_to_sort if not col.startswith("face")]

    # Now, sort these columns
    cols_to_sort.sort()
    # Now, create the final column order and rearrange the DataFrame
    final_col_order = ["Frame", "Time"] + cols_to_sort
    xyz_labelled = xyz_labelled[final_col_order]

    # trying a fix...
    xyz_labelled["Frame"] = xyz_labelled["Frame"].astype(int)

    # Get column names from dataframe
    columns = xyz_labelled.columns

    # Remove '_x', '_y', and '_z' suffixes and get unique names
    tracked_points = list(set([col.rsplit("_", 1)[0] for col in columns if col.endswith(("_x", "_y", "_z"))]))
    tracked_points.sort()

    num_markers = len(tracked_points)
    data_rate = int(mean_frame_rate)
    units = "m"
    # original_data_rate = int(mean_frame_rate)
    orig_data_start_frame = 0
    num_frames = len(xyz_labelled) - 1

    trc_path = Path(target_path.parent, f"{target_path.stem}.trc")
    trc_filename = str(trc_path)

    # this will create the formatted .trc file
    with open(trc_path, "wt", newline="", encoding="utf-8") as out_file:
        tsv_writer = csv.writer(out_file, delimiter="\t")
        tsv_writer.writerow(["PathFileType", "4", "(X/Y/Z)", trc_filename])
        tsv_writer.writerow(
            [
                "DataRate",
                "CameraRate",
                "NumFrames",
                "NumMarkers",
                "Units",
                "OrigDataRate",
                "OrigDataStartFrame",
                "OrigNumFrames",
            ]
        )
        tsv_writer.writerow(
            [
                data_rate,
                int(mean_frame_rate),
                num_frames,
                num_markers,
                units,
                int(mean_frame_rate),
                orig_data_start_frame,
                num_frames,
            ]
        )

        # create names of trajectories, skipping two columns (top of table)
        header_names = ["Frame#", "Time"]
        for trajectory in tracked_points:
            header_names.append(trajectory)
            header_names.append("")
            header_names.append("")

        tsv_writer.writerow(header_names)

        # create labels for x,y,z axes (below landmark names in header)
        header_names = ["", ""]
        for i in range(1, len(tracked_points) + 1):
            header_names.append("X" + str(i))
            header_names.append("Y" + str(i))
            header_names.append("Z" + str(i))

        tsv_writer.writerow(header_names)

        # the .trc fileformat expects a blank fourth line
        tsv_writer.writerow("")

        # this is here primarily for testing purposes right now..
        # filLs None with zeros.
        # df_xyz_labelled.fillna(0,inplace=True)

        # and finally actually write the trajectories
        for row in range(0, len(xyz_labelled)):
            row_data = xyz_labelled.iloc[row].tolist()

            # Convert the 'Frame' column value to int to satisfy trc format requirements
            frame_index = xyz_labelled.columns.get_loc("Frame")
            row_data[frame_index] = int(row_data[frame_index])

            tsv_writer.writerow(row_data)



import os
import subprocess
import sys

from PySide6.QtCore import QFileSystemWatcher, Qt
from PySide6.QtWidgets import QGridLayout, QHBoxLayout, QLabel, QPushButton, QSpinBox, QTextBrowser, QWidget

import caliscope.logger
from caliscope.controller import Controller
from caliscope.gui.synched_frames_display import SynchedFramesDisplay
from caliscope.gui.utils.spinbox_utils import setup_spinbox_sizing

logger = caliscope.logger.get(__name__)


class WorkspaceSummaryWidget(QWidget):
    def __init__(self, controller: Controller):
        super().__init__()

        self.controller = controller
        self.watcher = QFileSystemWatcher()

        # self.directory = QLabel(str(self.controller.workspace))
        self.open_workspace_folder_btn = QPushButton("Open Workspace Folder", self)
        self.calibrate_btn = QPushButton("Calibrate Capture Volume", self)
        self.reload_workspace_btn = QPushButton("Reload Workspace")

        self.camera_count_spin = QSpinBox()
        self.camera_count_spin.setValue(self.controller.get_camera_count())
        setup_spinbox_sizing(self.camera_count_spin,min_value=1,max_value=100,padding=30)
        self.camera_count_spin.setMaximumWidth(40)

        self.status_HTML = QTextBrowser()
        # Set the layout for the widget
        self.place_widgets()
        self.connect_widgets()

        self.update_status()

    def place_widgets(self):
        # Layout
        self.layout = QGridLayout()
        self.setLayout(self.layout)
        self.layout.addWidget(self.status_HTML, 0, 0, 1, 4)

        camera_spin_layout = QHBoxLayout()
        camera_spin_layout.addWidget(QLabel("Cameras:"), alignment=Qt.AlignmentFlag.AlignRight)
        camera_spin_layout.addWidget(self.camera_count_spin, alignment=Qt.AlignmentFlag.AlignLeft)
        self.layout.addLayout(
            camera_spin_layout,
            1,
            0,
        )
        self.layout.addWidget(self.reload_workspace_btn, 1, 1)
        self.layout.addWidget(self.open_workspace_folder_btn, 1, 2)
        self.layout.addWidget(self.calibrate_btn, 1, 3)

    def connect_widgets(self):
        self.open_workspace_folder_btn.clicked.connect(self.open_workspace)
        self.calibrate_btn.clicked.connect(self.on_calibrate_btn_clicked)
        self.camera_count_spin.valueChanged.connect(self.set_camera_count)
        self.controller.show_synched_frames.connect(self.show_synched_frames)

    def on_calibrate_btn_clicked(self):
        logger.info("Calling controller to process extrinsic streams into 2D data")
        # Call the extrinsic calibration method in the controller
        self.controller.calibrate_capture_volume()

    def set_camera_count(self, value):
        self.controller.set_camera_count(value)

    def open_workspace(self):
        logger.info(f"Opening workspace within File Explorer...  located at {self.controller.workspace}")
        if sys.platform == "win32":
            os.startfile(self.controller.workspace)
        elif sys.platform == "darwin":
            subprocess.run(["open", self.controller.workspace])
        else:  # Linux and Unix-like systems
            subprocess.run(["xdg-open", self.controller.workspace])

    def show_synched_frames(self):
        logger.info("About to launch synced Frames Display")
        self.display_window = SynchedFramesDisplay(self.controller.extrinsic_stream_manager)
        self.display_window.show()

    def update_status(self):
        self.status_HTML.setHtml(self.controller.workspace_guide.get_html_summary())



# logger.setLevel(logging.DEBUG)
import time
from queue import Queue
from threading import Event, Thread

import numpy as np

import caliscope.logger
from caliscope.packets import SyncPacket

logger = caliscope.logger.get(__name__)

DROPPED_FRAME_TRACK_WINDOW = 100  # trailing frames tracked for reporting purposes


class Synchronizer:
    def __init__(self, streams: dict):
        self.streams = streams
        self.current_synched_frames = None

        self.synched_frames_subscribers = []  # queues that will receive actual frame data

        self.all_frame_packets = {}
        self.stop_event = Event()
        self.frames_complete = False  # only relevant for video playback, but provides a way to wrap up the thread

        self.ports = []
        self.frame_packet_queues = {}
        for port, stream in self.streams.items():
            self.ports.append(port)
            q = Queue(-1)
            self.frame_packet_queues[port] = q

        self.subscribed_to_streams = False  # not subscribed yet
        self.subscribe_to_streams()

        # note that self.fps target is set in set_stream_fps
        # self.set_stream_fps(fps_target)
        # self.fps_mean = fps_target

        # place to store a recent history of dropped frames
        self.dropped_frame_history = {port: [] for port in sorted(self.ports)}

        self.initialize_ledgers()
        self.start()

    def set_tracking_on_streams(self, track: bool):
        for port, stream in self.streams.items():
            stream.set_tracking_on(track)

    def update_dropped_frame_history(self):
        current_dropped: dict = self.current_sync_packet.dropped

        for port, dropped in current_dropped.items():
            self.dropped_frame_history[port].append(dropped)
            self.dropped_frame_history[port] = self.dropped_frame_history[port][-DROPPED_FRAME_TRACK_WINDOW:]

    @property
    def dropped_fps(self):
        """
        Averages dropped frame count across the observed history
        """
        return {port: np.mean(drop_history) for port, drop_history in self.dropped_frame_history.items()}

    # def set_stream_fps(self, fps_target):
    #     self.fps_target = fps_target
    #     logger.info(f"Attempting to change target fps in streams to {fps_target}")
    #     for port, stream in self.streams.items():
    #         stream.set_fps_target(fps_target)

    def subscribe_to_streams(self):
        for port, stream in self.streams.items():
            logger.info(f"Subscribing synchronizer to stream from port {port}")
            stream.subscribe(self.frame_packet_queues[port])
        self.subscribed_to_streams = True

    def unsubscribe_from_streams(self):
        for port, stream in self.streams.items():
            logger.info(f"unsubscribe synchronizer from port {port}")
            stream.unsubscribe(self.frame_packet_queues[port])
        self.subscribed_to_streams = False

    def stop(self):
        self.stop_event.set()
        self.thread.join()
        for t in self.threads:
            t.join()

    def initialize_ledgers(self):
        self.port_frame_count = {port: 0 for port in self.ports}
        self.port_current_frame = {port: 0 for port in self.ports}
        self.mean_frame_times = []

    def start(self):
        logger.info("About to submit Threadpool of frame Harvesters")
        self.threads = []
        for port, stream in self.streams.items():
            t = Thread(target=self.harvest_frame_packets, args=(stream,), daemon=True)
            t.start()
            self.threads.append(t)
        logger.info("Frame harvesters just submitted")

        logger.info("Starting frame synchronizer...")
        self.thread = Thread(target=self.synch_frames_worker, args=(), daemon=True)
        self.thread.start()

    def subscribe_to_sync_packets(self, q):
        logger.info("Adding queue to receive synched frames")
        self.synched_frames_subscribers.append(q)

    def release_sync_packet_q(self, q):
        logger.info("Releasing record queue")
        self.synched_frames_subscribers.remove(q)

    def harvest_frame_packets(self, stream):
        port = stream.port

        logger.info(f"Beginning to collect data generated at port {port}")

        while not self.stop_event.is_set():
            frame_packet = self.frame_packet_queues[port].get()
            frame_index = self.port_frame_count[port]

            self.all_frame_packets[f"{port}_{frame_index}"] = frame_packet
            self.port_frame_count[port] += 1

            logger.debug(
                f"Frame data harvested from reel {frame_packet.port} with index {frame_index} and frame time of {frame_packet.frame_time}"  # noqa E501
            )

        logger.info(f"Frame harvester for port {port} completed")

    # get minimum value of frame_time for next layer
    def earliest_next_frame(self, port):
        """Looks at next unassigned frame across the ports to determine
        the earliest time at which each of them was read"""
        times_of_next_frames = []
        for p in self.ports:
            next_index = self.port_current_frame[p] + 1
            frame_data_key = f"{p}_{next_index}"

            # problem with outpacing the threads reading data in, so wait if need be
            while frame_data_key not in self.all_frame_packets.keys():
                logger.debug(f"Waiting in a loop for frame data to populate with key: {frame_data_key}")
                if self.subscribed_to_streams:
                    time.sleep(0.1)
                else:
                    # provide infrequent updates of busy waiting
                    if int(time.time()) % 10 == 0:
                        logger.info("Synchronizer not subscribed to any streams and busy waiting...")
                    time.sleep(1)

            next_frame_time = self.all_frame_packets[frame_data_key].frame_time

            if next_frame_time == -1:
                logger.info(f"End of frames at port {p} detected; ending synchronization")

                self.frames_complete = True
                self.stop_event.set()

            if p != port:
                times_of_next_frames.append(next_frame_time)

        return min(times_of_next_frames)

    def latest_current_frame(self, port):
        """Provides the latest frame_time of the current frames not inclusive of the provided port"""
        times_of_current_frames = []
        for p in self.ports:
            current_index = self.port_current_frame[p]
            frame_data_key = f"{p}_{current_index}"
            current_frame_time = self.all_frame_packets[frame_data_key].frame_time
            if p != port:
                times_of_current_frames.append(current_frame_time)

        return max(times_of_current_frames)

    def frame_slack(self):
        """Determine how many unassigned frames are sitting in self.dataframe"""
        slack = [self.port_frame_count[port] - self.port_current_frame[port] for port in self.ports]
        logger.debug(f"Slack in frames is {slack}")
        return min(slack)

    def average_fps(self):
        if len(self.mean_frame_times) > 10:  # only looking at the most recent layers
            self.mean_frame_times = self.mean_frame_times[-10:]

        delta_t = np.diff(self.mean_frame_times)
        mean_delta_t = np.mean(delta_t)

        return 1 / mean_delta_t

    def synch_frames_worker(self):
        logger.info("Waiting for all ports to begin harvesting corners...")

        sync_index = 0

        logger.info("About to start synchronizing frames...")
        while not self.stop_event.is_set():
            current_frame_packets = {}

            layer_frame_times = []

            # build earliest next/latest current dictionaries for each port to determine where to put frames
            # must be done before going in and making any updates to the frame index
            earliest_next = {}
            latest_current = {}

            for port in self.ports:
                earliest_next[port] = self.earliest_next_frame(port)
                latest_current[port] = self.latest_current_frame(port)
                current_frame_index = self.port_current_frame[port]

            for port in self.ports:
                current_frame_index = self.port_current_frame[port]

                port_index_key = f"{port}_{current_frame_index}"
                current_frame_packet = self.all_frame_packets[port_index_key]
                frame_time = current_frame_packet.frame_time

                # don't put a frame in a synched frame packet if the next packet has a frame before it
                if frame_time > earliest_next[port]:
                    # definitly should be put in the next layer and not this one
                    current_frame_packets[port] = None
                    logger.warning(f"Skipped frame at port {port}: > earliest_next")
                elif (
                    earliest_next[port] - frame_time < frame_time - latest_current[port]
                ):  # frame time is closer to earliest next than latest current
                    # if it's closer to the earliest next frame than the latest current frame, bump it up
                    # only applying for 2 camera setup where I noticed this was an issue (frames stay out of synch)
                    current_frame_packets[port] = None
                    logger.warning(f"Skipped frame at port {port}: delta < time-latest_current")
                else:
                    # add the data and increment the index
                    current_frame_packets[port] = self.all_frame_packets.pop(port_index_key)
                    # frame_packets[port]["sync_index"] = sync_index
                    self.port_current_frame[port] += 1
                    layer_frame_times.append(frame_time)
                    logger.debug(
                        f"Adding to layer from port {port} at index {current_frame_index} and frame time: {frame_time}"
                    )

            logger.debug(f"Unassigned Frames: {len(self.all_frame_packets)}")

            self.mean_frame_times.append(np.mean(layer_frame_times))

            logger.debug(f"Updating sync packet for sync_index {sync_index}")
            self.current_sync_packet = SyncPacket(sync_index, current_frame_packets)

            self.update_dropped_frame_history()

            sync_index += 1

            if self.stop_event.is_set():
                logger.info("Sending `None` on queue to signal end of synced frames.")
                self.current_sync_packet = None

            for q in self.synched_frames_subscribers:
                q.put(self.current_sync_packet)
                if self.current_sync_packet is not None:
                    logger.debug(
                        f"Placing new synched frames packet on queue with {self.current_sync_packet.frame_packet_count} frames"  # noqa E501
                    )
                    logger.debug(f"Placing new synched frames with index {self.current_sync_packet.sync_index}")

                    # provide infrequent notice of synchronizer activity
                    if self.current_sync_packet.sync_index % 100 == 0:
                        logger.info(f"Placing new synched frames with index {self.current_sync_packet.sync_index}")
                else:
                    logger.info("signaling end of frames with `None` packet on subscriber queue.")
                    for port, q in self.frame_packet_queues.items():
                        logger.info(f"Currently {q.qsize()} frame packets unprocessed for port {port}")

            self.fps_mean = self.average_fps()

        logger.info("Frame synch worker successfully ended")



from time import time

import numpy as np
import pandas as pd
from numba import jit
from numba.typed import Dict, List

import caliscope.logger
from caliscope.cameras.camera_array import CameraArray, CameraData

logger = caliscope.logger.get(__name__)


# helper function to avoid use of np.unique(return_counts=True) which doesn't work with jit
@jit(nopython=True, cache=True)
def unique_with_counts(arr):
    sorted_arr = np.sort(arr)
    unique_values = [sorted_arr[0]]
    counts = [1]

    for i in range(1, len(sorted_arr)):
        if sorted_arr[i] != sorted_arr[i - 1]:
            unique_values.append(sorted_arr[i])
            counts.append(1)
        else:
            counts[-1] += 1

    return np.array(unique_values), np.array(counts)


#####################################################################################
# The following code is adapted from the `Anipose` project,
# in particular the `triangulate_simple` function of `aniposelib`
# Original author:  Lili Karashchuk
# Project: https://github.com/lambdaloop/aniposelib/
# Original Source Code : https://github.com/lambdaloop/aniposelib/blob/d03b485c4e178d7cff076e9fe1ac36837db49158/aniposelib/cameras.py#L21
# This code is licensed under the BSD 2-Clause License
# BSD 2-Clause License

# Copyright (c) 2019, Lili Karashchuk
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@jit(nopython=True, parallel=True, cache=True)
def triangulate_sync_index(
    projection_matrices: Dict, current_camera_indices: np.ndarray, current_point_id: np.ndarray, current_img: np.ndarray
):
    # sync_indices_xyz = List()
    point_indices_xyz = List()
    obj_xyz = List()

    unique_points, point_counts = unique_with_counts(current_point_id)
    for index in range(len(point_counts)):
        if point_counts[index] > 1:
            # triangulate that point...
            point = unique_points[index]
            points_xy = current_img[current_point_id == point]
            camera_ids = current_camera_indices[current_point_id == point]

            num_cams = len(camera_ids)
            A = np.zeros((num_cams * 2, 4))
            for i in range(num_cams):
                x, y = points_xy[i]
                P = projection_matrices[camera_ids[i]]
                A[(i * 2) : (i * 2 + 1)] = x * P[2] - P[0]
                A[(i * 2 + 1) : (i * 2 + 2)] = y * P[2] - P[1]
            u, s, vh = np.linalg.svd(A, full_matrices=True)
            point_xyzw = vh[-1]
            point_xyz = point_xyzw[:3] / point_xyzw[3]

            point_indices_xyz.append(point)
            obj_xyz.append(point_xyz)

    return point_indices_xyz, obj_xyz


# End of adapted code
##################################################################################


def triangulate_xy(xy: pd.DataFrame, camera_array: CameraArray) -> pd.DataFrame:
    """
    xy data comes in as viewed by the camera and it is undistorted as
    part of the triangulation process
    """
    # assemble numba compatible dictionary
    projection_matrices = camera_array.projection_matrices

    # Code here to undistort all image points
    undistorted_xy = undistort_batch(xy, camera_array)

    xyz = {
        "sync_index": [],
        "point_id": [],
        "x_coord": [],
        "y_coord": [],
        "z_coord": [],
    }

    sync_index_max = xy["sync_index"].max()

    start = time()
    last_log_update = int(start)  # only report progress each second

    logger.info("About to begin triangulation...due to jit, first round of calculations may take a moment.")
    for index in xy["sync_index"].unique():
        active_index = xy["sync_index"] == index

        # load variables for given sync index
        port = xy["port"][active_index].to_numpy()
        point_ids = undistorted_xy["point_id"][active_index].to_numpy()
        img_loc_x = undistorted_xy["img_loc_undistort_x"][active_index].to_numpy()
        img_loc_y = undistorted_xy["img_loc_undistort_y"][active_index].to_numpy()
        raw_xy = np.vstack([img_loc_x, img_loc_y]).T

        # the fancy part
        point_id_xyz, points_xyz = triangulate_sync_index(projection_matrices, port, point_ids, raw_xy)

        if len(point_id_xyz) > 0:
            # there are points to store so store them...
            xyz["sync_index"].extend([index] * len(point_id_xyz))
            xyz["point_id"].extend(point_id_xyz)

            points_xyz = np.array(points_xyz)
            xyz["x_coord"].extend(points_xyz[:, 0].tolist())
            xyz["y_coord"].extend(points_xyz[:, 1].tolist())
            xyz["z_coord"].extend(points_xyz[:, 2].tolist())

        # only log percent complete each second
        if int(time()) - last_log_update >= 1:
            percent_complete = int(100 * (index / sync_index_max))
            logger.info(f"(Stage 2 of 2): Triangulation of (x,y) point estimates is {percent_complete}% complete")
            last_log_update = int(time())

    # convert to dataframe prior to returning
    xyz = pd.DataFrame(xyz)
    return xyz


def undistort(points, camera: CameraData, iter_num=3) -> np.ndarray:
    """
    points: (n,2) dimensional np.ndarray
    returns: (2,n) dimensional np.ndarray
    """

    # implementing a function described here: https://yangyushi.github.io/code/2020/03/04/opencv-undistort.html
    # supposedly a better implementation than OpenCV
    k1, k2, p1, p2, k3 = camera.distortions
    fx, fy = camera.matrix[0, 0], camera.matrix[1, 1]
    cx, cy = camera.matrix[:2, 2]

    x, y = points.T[0], points.T[1]

    x = (x - cx) / fx
    x0 = x
    y = (y - cy) / fy
    y0 = y

    for _ in range(iter_num):
        r2 = x**2 + y**2
        k_inv = 1 / (1 + k1 * r2 + k2 * r2**2 + k3 * r2**3)
        delta_x = 2 * p1 * x * y + p2 * (r2 + 2 * x**2)
        delta_y = p1 * (r2 + 2 * y**2) + 2 * p2 * x * y
        x = (x0 - delta_x) * k_inv
        y = (y0 - delta_y) * k_inv
    return np.array((x * fx + cx, y * fy + cy))


def undistort_batch(xy_df: pd.DataFrame, camera_array: CameraArray) -> pd.DataFrame:
    undistorted_points = []
    for port, camera in camera_array.cameras.items():
        logger.info(f"Processing points from camera {port}")
        subset_xy = xy_df.query(f"port == {port}").copy()
        points = np.vstack([subset_xy["img_loc_x"], subset_xy["img_loc_y"]]).T
        x, y = undistort(points, camera)
        subset_xy["img_loc_undistort_x"] = x
        subset_xy["img_loc_undistort_y"] = y
        undistorted_points.append(subset_xy)

    logger.info("Assembling undistorted dataframe")

    xy_undistorted_df = pd.concat(undistorted_points)
    return xy_undistorted_df



# %%


from dataclasses import dataclass
from itertools import permutations
from pathlib import Path

import numpy as np
import rtoml

import caliscope.logger
from caliscope import __root__
from caliscope.calibration.capture_volume.capture_volume import CaptureVolume
from caliscope.calibration.capture_volume.helper_functions.get_point_estimates import (
    get_point_estimates,
)
from caliscope.calibration.capture_volume.point_estimates import PointEstimates
from caliscope.cameras.camera_array import CameraArray, CameraData

logger = caliscope.logger.get(__name__)


@dataclass
class StereoPair:
    """
    A dataclass to hold the extrinsic parameters associated with the cv2.stereoCalibrate
    function output. Additionally provides some convenience methods to get common transformations
    of the data.

    From the first output of cv2.stereocalibrate, variations can be created by inverting camera
    relationships, and chaining together relative camera positions so that all possible pairs
    of cameras are represented. This dataclass is the building block of that larger process.
    """

    primary_port: int
    secondary_port: int
    error_score: float  # when chaining operations, this will be a cumulative number
    translation: np.ndarray
    rotation: np.ndarray

    @property
    def pair(self):
        return (self.primary_port, self.secondary_port)

    @property
    def transformation(self):
        R_stack = np.vstack([self.rotation, np.array([0, 0, 0])])
        t_stack = np.vstack([self.translation, np.array([1])])
        Tranformation = np.hstack([R_stack, t_stack])
        return Tranformation


def get_inverted_stereopair(stereo_pair: StereoPair) -> StereoPair:
    primary_port = stereo_pair.secondary_port
    secondary_port = stereo_pair.primary_port
    error_score = stereo_pair.error_score

    inverted_transformation = np.linalg.inv(stereo_pair.transformation)
    rotation = inverted_transformation[0:3, 0:3]
    translation = inverted_transformation[0:3, 3:]

    inverted_stereopair = StereoPair(
        primary_port=primary_port,
        secondary_port=secondary_port,
        error_score=error_score,
        translation=translation,
        rotation=rotation,
    )
    return inverted_stereopair


def get_bridged_stereopair(pair_A_B: StereoPair, pair_B_C: StereoPair) -> StereoPair:
    port_A = pair_A_B.primary_port
    port_C = pair_B_C.secondary_port

    A_B_error = pair_A_B.error_score
    B_C_error = pair_B_C.error_score
    A_C_error = A_B_error + B_C_error

    # new transformations are added on the left
    # https://youtube.com/watch?v=q0mRtuiKSKg&feature=shares&t=66
    bridged_transformation = np.matmul(pair_B_C.transformation, pair_A_B.transformation)
    bridged_rotation = bridged_transformation[0:3, 0:3]
    bridged_translation = bridged_transformation[None, 0:3, 3].T

    stereo_A_C = StereoPair(
        primary_port=port_A,
        secondary_port=port_C,
        error_score=A_C_error,
        translation=bridged_translation,
        rotation=bridged_rotation,
    )

    return stereo_A_C


class CameraArrayInitializer:
    def __init__(self, config_path: Path):
        logger.info("Creating initial estimate of camera array based on stereopairs...")

        self.config = rtoml.load(config_path)
        self.ports = self._get_ports()
        self.estimated_stereopairs = self._get_captured_stereopairs()
        self._fill_stereopair_gaps()
        # self.best_camera_array = self.get_best_camera_array()

    def _fill_stereopair_gaps(self):
        """
        Loop across missing pairs and create bridged stereopairs when possible.
        It may be that one iteration is not sufficient to fill all missing pairs,
        so iterate until no more missing pairs...

        The code below uses a naming convention to describe the relationship between
        two stereo pairs (A,X) and (X,C) that can be used to build a bridge stereopair (A,C)
        """

        # fill with dummy value to get the loop running
        missing_count_last_cycle = -1

        while len(self._get_missing_stereopairs()) != missing_count_last_cycle:
            # prep the variable. if it doesn't go down, terminate
            missing_count_last_cycle = len(self._get_missing_stereopairs())

            for pair in self._get_missing_stereopairs():
                port_A = pair[0]
                port_C = pair[1]

                # get lists of all the estimiated stereopairs that might bridge across test_missing
                all_pairs_A_X = [pair for pair in self.estimated_stereopairs.keys() if pair[0] == port_A]
                all_pairs_X_C = [pair for pair in self.estimated_stereopairs.keys() if pair[1] == port_C]

                stereopair_A_C = None

                for pair_A_X in all_pairs_A_X:
                    for pair_X_C in all_pairs_X_C:
                        if pair_A_X[1] == pair_X_C[0]:
                            # A bridge can be formed!
                            stereopair_A_X = self.estimated_stereopairs[pair_A_X]
                            stereopair_X_C = self.estimated_stereopairs[pair_X_C]
                            possible_stereopair_A_C = get_bridged_stereopair(stereopair_A_X, stereopair_X_C)
                            if stereopair_A_C is None:
                                # current possibility is better than nothing
                                stereopair_A_C = possible_stereopair_A_C
                            else:
                                # check if it's better than what you have already
                                # if it is, then overwrite the old one
                                if stereopair_A_C.error_score > possible_stereopair_A_C.error_score:
                                    stereopair_A_C = possible_stereopair_A_C

                if stereopair_A_C is not None:
                    self.add_stereopair(stereopair_A_C)

        if len(self._get_missing_stereopairs()) > 0:
            raise ValueError("Insufficient stereopairs to allow array to be estimated")

    def _get_missing_stereopairs(self):
        possible_stereopairs = [pair for pair in permutations(self.ports, 2)]
        missing_stereopairs = [pair for pair in possible_stereopairs if pair not in self.estimated_stereopairs.keys()]

        return missing_stereopairs

    def _get_ports(self) -> list:
        ports = []
        for key, params in self.config.items():
            if key.split("_")[0] == "stereo":
                port_A = int(key.split("_")[1])
                port_B = int(key.split("_")[2])
                ports.append(port_A)
                ports.append(port_B)

        # convert to a unique list
        ports = list(set(ports))
        return ports

    def _get_captured_stereopairs(self) -> dict:
        stereopairs = {}

        # Create StereoPair objects for each saved stereocalibration output in config
        # this are maintained in a dictionary keyed off of the pair tuple
        for key, params in self.config.items():
            if key.split("_")[0] == "stereo":
                port_A = int(key.split("_")[1])
                port_B = int(key.split("_")[2])

                rotation = np.array(params["rotation"], dtype=np.float64)
                translation = np.array(params["translation"], dtype=np.float64)
                error = float(params["RMSE"])

                new_stereopair = StereoPair(
                    primary_port=port_A,
                    secondary_port=port_B,
                    error_score=error,
                    translation=translation,
                    rotation=rotation,
                )

                stereopairs[new_stereopair.pair] = new_stereopair

        # create another dictionary that will contain the inverted versions of the StereoPairs
        inverted_stereopairs = {}
        for pair, stereopair in stereopairs.items():
            a, b = pair
            inverted_pair = (b, a)
            inverted_stereopairs[inverted_pair] = get_inverted_stereopair(stereopair)

        # combine the dictionaries
        merged_stereopairs = {**stereopairs, **inverted_stereopairs}
        return merged_stereopairs

    def _get_scored_anchored_array(self, anchor_port: int) -> tuple:
        """
        Constructs a complete camera array based on the available stereopairs in
        self.all_stereopairs

        two return values:

            total_error_score: the sum of the error_scores of all stereopairs used in the
                            construction of the array

            camera_array: a CameraArray object anchored at the provided port
        """
        cameras = {}
        total_error_score = 0

        for key, data in self.config.items():
            if key.startswith("cam_"):  # and not self.config[key]["ignore"]:
                port = data["port"]
                size = data["size"]
                rotation_count = data["rotation_count"]
                error = data["error"]
                matrix = np.array(data["matrix"], dtype=np.float64)
                distortions = np.array(data["distortions"], dtype=np.float64)
                grid_count = data["grid_count"]

                # update with extrinsics, though place anchor camera at origin
                if port == anchor_port:
                    translation = np.array([0, 0, 0], dtype=np.float64).T
                    rotation = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float64)
                else:
                    anchored_stereopair = self.estimated_stereopairs[(anchor_port, port)]
                    translation = anchored_stereopair.translation[:, 0]
                    rotation = anchored_stereopair.rotation
                    total_error_score += anchored_stereopair.error_score

                cam_data = CameraData(
                    port=port,
                    size=size,
                    rotation_count=rotation_count,
                    error=error,
                    matrix=matrix,
                    distortions=distortions,
                    # exposure,
                    grid_count=grid_count,
                    # ignore=ignore,
                    # verified_resolutions,
                    translation=translation,
                    rotation=rotation,
                )

                cameras[port] = cam_data

        camera_array = CameraArray(cameras)

        return total_error_score, camera_array

    def get_best_camera_array(self):
        """
        returns the anchored camera array with the lowest total error score.
        Note that total error score is just a sum of individual errors for tracking
        and comparison purposes and does not have any signifigence in the context
        of reprojection error

        """
        array_error_scores = {}
        camera_arrays = {}
        # get the score for the anchored_stereopairs
        for port in self.ports:
            array_error_score, camera_array = self._get_scored_anchored_array(port)
            array_error_scores[port] = array_error_score
            camera_arrays[port] = camera_array

        best_anchor = min(array_error_scores, key=array_error_scores.get)

        best_initial_array = camera_arrays[best_anchor]

        return best_initial_array

    def add_stereopair(self, stereopair: StereoPair):
        self.estimated_stereopairs[stereopair.pair] = stereopair
        inverted_stereopair = get_inverted_stereopair(stereopair)
        self.estimated_stereopairs[inverted_stereopair.pair] = inverted_stereopair


# def get_anchored_pairs(anchor: int, all_stereopairs:dict)->dict:


if __name__ == "__main__":
    session_directory = Path(__root__, "tests", "sessions", "217")

    config_path = Path(session_directory, "config.toml")

    initializer = CameraArrayInitializer(config_path)

    camera_array = initializer.get_best_camera_array()

    extrinsic_calibration_xy = Path(session_directory, "point_data.csv")

    point_estimates: PointEstimates = get_point_estimates(camera_array, extrinsic_calibration_xy)

    capture_volume = CaptureVolume(camera_array, point_estimates)

    pair_A_B = initializer.estimated_stereopairs[(0, 1)]
    pair_B_C = initializer.estimated_stereopairs[(1, 2)]

    bridged_pair = get_bridged_stereopair(pair_A_B, pair_B_C)
    logger.info(bridged_pair)

    # capture_volume.optimize()


# %%



import sys
from pathlib import Path

from PySide6.QtWidgets import (
    QApplication,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

import caliscope.logger
from caliscope.controller import Controller
from caliscope.gui.camera_management.playback_widget import (
    IntrinsicCalibrationWidget,
)

logger = caliscope.logger.get(__name__)


class MultiIntrinsicPlaybackWidget(QWidget):
    def __init__(self, controller):
        super().__init__()
        self.controller = controller
        self.initUI()

    def initUI(self):
        self.tabWidget = QTabWidget(self)
        self.tabWidget.setTabPosition(QTabWidget.South)  # Tabs on the left side
        self.loadTabs()

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.tabWidget)
        # self.resize(1200, 800)

    def loadTabs(self):
        logger.info("Beginning to load individual tabs")
        for camera in self.controller.camera_array.cameras.values():
            logger.info(f"About to create calibration widget for camera {camera.port}")
            tab = IntrinsicCalibrationWidget(controller=self.controller, port=camera.port)
            logger.info(f"Calibration widget for camera {camera.port} successfully created")
            self.tabWidget.addTab(tab, f"Cam {camera.port}")


if __name__ == "__main__":
    app = QApplication(sys.argv)

    workspace_dir = Path(r"C:\Users\Mac Prible\OneDrive\caliscope\prerecorded_workflow")
    controller = Controller(workspace_dir)
    controller.load_camera_array()
    controller.load_intrinsic_stream_manager()

    mainWin = MultiIntrinsicPlaybackWidget(controller)
    mainWin.show()
    sys.exit(app.exec())



# %%

import pickle
import sys
from pathlib import Path

import cv2
import numpy as np
import scipy
from PySide6.QtWidgets import QApplication

import caliscope.logger
from caliscope import __root__
from caliscope.calibration.capture_volume.helper_functions.get_point_estimates import (
    get_point_estimates,
)
from caliscope.calibration.capture_volume.point_estimates import PointEstimates
from caliscope.calibration.charuco import Charuco
from caliscope.cameras.camera_array import CameraArray, CameraData

logger = caliscope.logger.get(__name__)


# Proceeding with basic idea that these functions will go into CaptureVolume.
# so use capture_volume here which may just become self later on.


def get_world_corners_xyz(point_estimates: PointEstimates, sync_index: int) -> np.ndarray:
    """
    returns the estimated x,y,z position of the board corners at the given sync index
    note that the array is ordered according to the charuco id
    """
    sync_indices = point_estimates.sync_indices  # convienent shortening

    charuco_ids = point_estimates.point_id[sync_indices == sync_index]
    unique_charuco_id = np.unique(charuco_ids)
    unique_charuco_id.sort()

    # pull out the 3d point estimate indexes associated with the chosen sync_index
    # note that this will include duplicates
    obj_indices = point_estimates.obj_indices[sync_indices == sync_index]
    # now get the actual x,y,z estimate associated with these unique charucos
    obj_xyz = point_estimates.obj[obj_indices]

    sorter = np.argsort(charuco_ids)
    # need to get charuco ids associated with the 3 point positions
    unique_charuco_xyz_index = sorter[np.searchsorted(charuco_ids, unique_charuco_id, sorter=sorter)]
    world_corners_xyz = obj_xyz[unique_charuco_xyz_index]
    return world_corners_xyz


def get_board_corners_xyz(point_estimates: PointEstimates, sync_index: int, charuco: Charuco) -> np.ndarray:
    """
    Returns corner positions in board world (x,y,0) for the corners with estimated point
    coordinates at the give sync_index
    note that the array is ordered according to the charuco id
    """
    sync_indices = point_estimates.sync_indices  # convienent shortening
    charuco_ids = point_estimates.point_id[sync_indices == sync_index]
    unique_charuco_id = np.unique(charuco_ids)
    unique_charuco_id.sort()

    board_corners_xyz = charuco.board.getChessboardCorners()[unique_charuco_id]
    return board_corners_xyz


def get_anchor_cameras(camera_array: CameraArray, point_estimates: PointEstimates, sync_index: int) -> list:
    """
    Returns the camera data objects that have an actual view of the board
    at the sync index and therefore can be used to estimate the pose from pnp

    Weird results will happen if the camera has its back turned to the board and
    the corners are projected into it.

    """
    sync_indices = point_estimates.sync_indices  # convienent shortening
    camera_views = point_estimates.camera_indices[sync_indices == sync_index]
    camera_ports, camera_counts = np.unique(camera_views, return_counts=True)

    anchor_cameras = []
    for port in camera_ports:
        cam: CameraData = camera_array.cameras[port]
        anchor_cameras.append(cam)

    return anchor_cameras


def get_rvec_tvec_from_board_pose(
    camera_array: CameraArray,
    point_estimates: PointEstimates,
    sync_index: int,
    charuco: Charuco,
):
    world_corners_xyz = get_world_corners_xyz(point_estimates, sync_index)
    board_corners_xyz = get_board_corners_xyz(point_estimates, sync_index, charuco)
    anchor_cameras = get_anchor_cameras(camera_array, point_estimates, sync_index)

    rvecs = []
    tvecs = []

    for camera in anchor_cameras:
        charuco_image_points, jacobian = cv2.projectPoints(
            world_corners_xyz,
            rvec=camera.rotation,
            tvec=camera.translation,
            cameraMatrix=camera.matrix,
            distCoeffs=np.array([0, 0, 0, 0, 0], dtype=np.float32),  # because points are via bundle adj., no distortion
        )

        # use solvepnp to estimate the pose of the camera relative to the board
        # this provides a good estimate of rotation, but not of translation
        retval, rvec, tvec = cv2.solvePnP(
            board_corners_xyz,
            charuco_image_points,
            cameraMatrix=camera.matrix,
            distCoeffs=np.array([0, 0, 0, 0, 0], dtype=np.float32),
        )

        anchor_board_transform = rvec_tvec_to_transform(rvec, tvec)

        # back into the shift in the world change of origin implied by the
        # pose of the camera relative to the board given its previous
        # pose in the old frame of reference
        origin_shift_transform = np.matmul(np.linalg.inv(camera.transformation), anchor_board_transform)

        rvec, tvec = transform_to_rvec_tvec(origin_shift_transform)

        rvecs.append(rvec)
        tvecs.append(tvec)

    mean_rvec = mean_vec(rvecs)
    mean_tvec = mean_vec(tvecs)

    # get mean values
    return mean_rvec, mean_tvec


def mean_vec(vecs):
    hstacked_vec = None

    for vec in vecs:
        if hstacked_vec is None:
            hstacked_vec = vec
        else:
            hstacked_vec = np.hstack([hstacked_vec, vec])

    mean_vec = np.mean(hstacked_vec, axis=1)
    mean_vec = np.expand_dims(mean_vec, axis=1)

    return mean_vec


def transform_to_rvec_tvec(transformation: np.ndarray):
    rot_matrix = transformation[0:3, 0:3]
    rvec = cv2.Rodrigues(rot_matrix)[0]
    tvec = np.expand_dims(transformation[0:3, 3], axis=1)
    return rvec, tvec


def rvec_tvec_to_transform(rvec: np.ndarray, tvec: np.ndarray) -> np.ndarray:
    # might send a rotation matrix into here so check
    # that rvec is a rodrigues vector before converting
    if len(rvec.shape) == 1 or rvec.shape[1] == 1:
        rvec = cv2.Rodrigues(rvec)[0]

    # if tvec doesn't have the extra dimension, add it
    if len(tvec.shape) == 1:
        tvec = np.expand_dims(tvec, axis=1)

    transform = np.hstack([rvec, tvec])
    transform = np.vstack([transform, np.array([0, 0, 0, 1], np.float32)])
    return transform


def world_board_distance(tvec_xyz: np.ndarray, good_rvec: np.ndarray, raw_world_xyz, board_corners_xyz):
    scale = np.expand_dims(np.ones(raw_world_xyz.shape[0]), 1)
    raw_world_xyzh = np.hstack([raw_world_xyz, scale])

    origin_shift_transform = rvec_tvec_to_transform(good_rvec, tvec_xyz)

    new_origin_xyzh = np.matmul(np.linalg.inv(origin_shift_transform), raw_world_xyzh.T).T
    new_origin_xyz = new_origin_xyzh[:, 0:3]

    delta_xyz = new_origin_xyz - board_corners_xyz
    logger.info(f"Delta_xyz is {delta_xyz}")
    return delta_xyz.ravel()


def get_board_origin_transform(
    camera_array: CameraArray,
    point_estimates: PointEstimates,
    sync_index: int,
    charuco: Charuco,
):
    """
    Returns the 4x4 transformation matrix that will shift the capture volumes camera and
    point data to be represented with an origin that aligns with the origin of the charuco board.

    This is the primary function that is exposed from this file to the CaptureVolume class
    """
    # get initial approximation of the transformation to apply
    good_rvec, poor_tvec = get_rvec_tvec_from_board_pose(
        camera_array=camera_array,
        point_estimates=point_estimates,
        sync_index=sync_index,
        charuco=charuco,
    )

    world_board = get_world_corners_xyz(point_estimates, sync_index)
    target_board = get_board_corners_xyz(point_estimates, sync_index, charuco)
    initial_tvec = poor_tvec[:, 0]  # prep for least_squares array

    least_sq_result = scipy.optimize.least_squares(
        world_board_distance, initial_tvec, args=(good_rvec, world_board, target_board)
    )

    optimal_tvec = least_sq_result.x
    final_transform = rvec_tvec_to_transform(good_rvec, optimal_tvec)

    return final_transform


if __name__ == "__main__":
    #
    from caliscope.calibration.capture_volume.capture_volume import CaptureVolume
    from caliscope.cameras.camera_array_initializer import CameraArrayInitializer
    from caliscope.gui.vizualize.calibration.capture_volume_visualizer import (
        CaptureVolumeVisualizer,
    )
    from caliscope.gui.vizualize.calibration.capture_volume_widget import (
        CaptureVolumeWidget,
    )
    from caliscope.session.session import LiveSession

    # test_scenario = "4_cameras_nonoverlap"
    # test_scenario = "3_cameras_middle"
    # test_scenario = "2_cameras_linear"
    # test_scenario = "3_cameras_triangular"
    test_scenario = "4_cameras_beginning"  # initial translation off
    # test_scenario = "3_cameras_midlinear"

    anchor_camera_override = None
    # anchor_camera_override = 2

    origin_sync_indices = {
        "4_cameras_nonoverlap": 23,
        "2_cameras_linear": 77,
        "3_cameras_middle": 20,
        "4_cameras_beginning": 230,
        "3_cameras_triangular": 25,
        "3_cameras_midlinear": 14,
    }

    session_directory = Path(__root__, "tests", test_scenario)
    point_data_csv_path = Path(session_directory, "point_data.csv")
    config_path = Path(session_directory, "config.toml")

    # need to get the charuco board that was used during the session for later
    session = LiveSession(session_directory)
    tracker = session.charuco

    REOPTIMIZE_CAPTURE_VOLUME = True
    # REOPTIMIZE_CAPTURE_VOLUME = False

    if REOPTIMIZE_CAPTURE_VOLUME:
        array_initializer = CameraArrayInitializer(config_path)
        camera_array = array_initializer.get_best_camera_array()
        point_estimates = get_point_estimates(camera_array, point_data_csv_path)

        print("Optimizing initial camera array configuration ")

        capture_volume = CaptureVolume(camera_array, point_estimates)
        capture_volume._save(session_directory, "initial")
        capture_volume.optimize()
        capture_volume._save(session_directory, "optimized")
    else:
        saved_CV_path = Path(session_directory, "capture_volume_stage_1_optimized.pkl")
        with open(saved_CV_path, "rb") as f:
            capture_volume: CaptureVolume = pickle.load(f)

    origin_sync_index = origin_sync_indices[test_scenario]
    logger.warning(f"New test sync index is {origin_sync_index}")

    capture_volume.set_origin_to_board(origin_sync_index, tracker)

    app = QApplication(sys.argv)
    vizr = CaptureVolumeVisualizer(capture_volume=capture_volume)

    vizr_dialog = CaptureVolumeWidget(vizr)
    vizr_dialog.show()

    sys.exit(app.exec())



from queue import Queue
from threading import Event

import numpy as np
from PySide6.QtCore import Qt, QThread, Signal
from PySide6.QtGui import QPixmap

import caliscope.logger
from caliscope.cameras.camera_array import CameraData
from caliscope.cameras.synchronizer import Synchronizer
from caliscope.gui.frame_emitters.tools import apply_rotation, cv2_to_qlabel, resize_to_square

logger = caliscope.logger.get(__name__)


class FrameDictionaryEmitter(QThread):
    # establish signals that will be displayed within the GUI
    FramesBroadcast = Signal(dict)
    # dropped_fps = Signal(dict)
    close_window = Signal()
    # GridCountBroadcast = Signal(int)
    # FrameIndexBroadcast = Signal(int, int)

    def __init__(self, synchronizer: Synchronizer, all_camera_data: dict[CameraData], pixmap_edge_length=500):
        # pixmap_edge length is from the display window. Keep the display area
        # square to keep life simple.
        super(FrameDictionaryEmitter, self).__init__()

        self.synchronizer = synchronizer
        self.streams = self.synchronizer.streams
        self.all_camera_data = all_camera_data

        self.sync_packet_q = Queue()
        self.synchronizer.subscribe_to_sync_packets(self.sync_packet_q)
        self.pixmap_edge_length = pixmap_edge_length
        self.keep_collecting = Event()
        logger.info("frame dictionary emitter initialized")

    def run(self):
        logger.info("Frame dictionary emitter beginning to run")
        self.keep_collecting.set()

        while self.keep_collecting.is_set():
            # Grab a frame from the queue and broadcast to displays
            # self.monocalibrator.grid_frame_ready_q.get()
            logger.debug("Getting sync packet from queue")
            sync_packet = self.sync_packet_q.get()
            if sync_packet is None:
                logger.info("End of playback signalled by synchronizer")
                break

            logger.debug(f"Sync packet: {sync_packet}")
            emitted_dict = {}
            for port, frame_packet in sync_packet.frame_packets.items():
                if frame_packet is None:
                    logger.info("plugging blank frame data")
                    frame = np.zeros((self.pixmap_edge_length, self.pixmap_edge_length, 3), dtype=np.uint8)
                else:
                    frame = frame_packet.frame_with_points

                rotation_count = self.streams[port].rotation_count
                frame = resize_to_square(frame)
                frame = apply_rotation(frame, rotation_count)
                image = cv2_to_qlabel(frame)
                pixmap = QPixmap.fromImage(image)

                if self.pixmap_edge_length:
                    pixmap = pixmap.scaled(
                        int(self.pixmap_edge_length),
                        int(self.pixmap_edge_length),
                        Qt.AspectRatioMode.KeepAspectRatio,
                    )

                emitted_dict[str(port)] = pixmap

            logger.debug(f"About to emit q_image_dict: {emitted_dict}")
            self.FramesBroadcast.emit(emitted_dict)

        self.close_window.emit()
        logger.info("Thread loop within dictionary frame emitter successfully ended")

    def stop(self):
        self.keep_collecting = False
        self.quit()



from itertools import combinations
from pathlib import Path

import cv2
import numpy as np
import pandas as pd
import rtoml

import caliscope.logger

logger = caliscope.logger.get(__name__)


class StereoCalibrator:
    def __init__(
        self,
        config_path: Path,
        point_data_path: Path,
    ):
        self.config_path = config_path
        self.config = rtoml.load(config_path)

        self.ports = []
        # set ports keeping in mind that something may be flagged for ignore
        for key, value in self.config.items():
            if key[0:4] == "cam_":
                self.ports.append(int(key[4:]))

        # import point data, adding coverage regions to each port
        raw_point_data = pd.read_csv(point_data_path)
        self.all_point_data = self.points_with_coverage_region(raw_point_data)
        self.all_boards = self.get_boards_with_coverage()

        # self.ports = [int(key[4:]) for key in self.config.keys() if key[0:3] == "cam"]
        self.pairs = [(i, j) for i, j in combinations(self.ports, 2) if i < j]

    def points_with_coverage_region(self, point_data: pd.DataFrame):
        """
        Efficiently create coverage region strings for points.
        """
        # Extract unique combinations of sync_index, point_id, and port
        point_ports = point_data[['sync_index', 'point_id', 'port']].drop_duplicates()

        # Convert port to strings for easier handling
        point_ports['port_str'] = point_ports['port'].astype(str)

        # Group by sync_index and point_id to collect ports
        grouped = point_ports.groupby(['sync_index', 'point_id'])['port_str'].apply(
            lambda x: '_' + '_'.join(sorted(x)) + '_'
        ).reset_index(name='coverage_region')

        # Merge back with original data
        result = point_data.merge(grouped, on=['sync_index', 'point_id'], how='left')

        return result

    # def points_with_coverage_region(self, point_data: pd.DataFrame):
    #     """
    #     Pivot the port columns and assemble a new string field that will show all of the cameras that
    #     observed a given corner at a single sync index.
    #     """

    #     points_w_pivoted_ports = (
    #         point_data.filter(["sync_index", "point_id", "port"])
    #         .pivot(index=["sync_index", "point_id"], columns="port", values="port")
    #         .reset_index()
    #         .fillna("")
    #     )

    #     def get_coverage_region(row, ports):
    #         """
    #         returns a string of the format "_0_1_2" for points which were captured
    #         by cameras 0,1 and 2, etc...
    #         """
    #         text = ""
    #         for port in ports:
    #             label = row[port]
    #             if label != "":
    #                 label = str(int(label))
    #                 text = text + "_" + label

    #         text = text + "_"

    #         return text

    #     points_w_pivoted_ports["coverage_region"] = points_w_pivoted_ports.apply(
    #         get_coverage_region, axis=1, args=(self.ports,)
    #     )

    #     points_w_pivoted_ports = points_w_pivoted_ports.filter(["sync_index", "point_id", "coverage_region"])
    #     points_w_regions = point_data.merge(points_w_pivoted_ports, "left", ["sync_index", "point_id"])

    #     return points_w_regions

    def get_boards_with_coverage(self):
        """
        create a dataframe of all the boards, including the most prevalent
        coverage region for each board
        Begin by getting the total number of corners observed for each board
        splitting out into its own creation method just for readability during init
        """

        board_points = (
            self.all_point_data.filter(["port", "sync_index", "point_id"])
            .groupby(["port", "sync_index"])
            .agg({"point_id": "count"})
            .rename(columns={"point_id": "point_count"})
            .reset_index()
        )

        # find the primary region each board is in, then merge in total counts
        all_boards = (
            self.all_point_data.filter(["port", "sync_index", "point_id", "coverage_region"])
            .groupby(["port", "sync_index", "coverage_region"])
            .agg({"point_id": "count"})
            .reset_index()
            .rename(
                {"point_id": "point_count"},
                axis=1,
            )
            .sort_values(
                ["port", "sync_index", "point_count"],
                axis=0,
                ascending=[True, True, False],
            )
            .groupby(["port", "sync_index"])
            .first()
            .reset_index()
            .rename(columns={"coverage_region": "primary_coverage_region"})
            .drop("point_count", axis=1)  # no longer means anthing
            .merge(board_points, "left", ["port", "sync_index"])  # merge in a point_count that means something
        )

        return all_boards

    def get_stereopair_data(self, pair: tuple, boards_sampled: int, random_state=1):
        """
        Efficiently extract data for a stereo pair.
        """
        a, b = pair
        a_str, b_str = str(a), str(b)

        # Create vectorized masks for filtering
        in_region_a = self.all_point_data['coverage_region'].str.contains(f'_{a_str}_')
        in_region_b = self.all_point_data['coverage_region'].str.contains(f'_{b_str}_')
        in_port_pair = self.all_point_data['port'].isin([a, b])

        # Filter points that are in the shared region and in one of the cameras
        pair_points = self.all_point_data[in_region_a & in_region_b & in_port_pair].copy()

        if pair_points.empty:
            logger.info(f"For pair {pair} there are no shared points")
            return None

        # Count points per board and filter to boards with enough points
        board_counts = pair_points.groupby(['sync_index', 'port']).size().reset_index(name='point_count')
        valid_boards = board_counts[board_counts['point_count'] >= 6]

        # Filter to port a only (to avoid duplicates)
        valid_boards_a = valid_boards[valid_boards['port'] == a][['sync_index', 'point_count']]

        if valid_boards_a.empty:
            logger.info(f"For pair {pair} there are no boards with sufficient points")
            return None

        # Sample boards
        sample_size = min(len(valid_boards_a), boards_sampled)

        if sample_size > 0:
            logger.info(f"Assembling {sample_size} shared boards for pair {pair}")

            # Sample boards with weighting
            weights = valid_boards_a['point_count'] ** 2
            selected_boards = valid_boards_a.sample(
                n=sample_size,
                weights=weights,
                random_state=random_state
            )

            # Filter points to selected boards
            selected_points = pair_points[pair_points['sync_index'].isin(selected_boards['sync_index'])]
            return selected_points
        else:
            logger.info(f"For pair {pair} there are no shared boards")
            return None


    # def get_stereopair_data(self, pair: tuple, boards_sampled: int, random_state=1) -> pd.DataFrame or None:
    #     # convenience function to get the points that are in the overlap regions of the pairs
    #     def in_pair(row: int, pair: tuple):
    #         """
    #         Uses the coverage_region string generated previously to flag points that are in
    #         a shared region of the pair
    #         """
    #         a, b = pair
    #         port_check = row.port == a or row.port == b

    #         a, b = str(a), str(b)

    #         region_check = ("_" + a + "_") in row.coverage_region and ("_" + b + "_") in row.coverage_region
    #         return region_check and port_check

    #     # flag the points that belong to the pair overlap regions
    #     self.all_point_data["in_pair"] = self.all_point_data.apply(in_pair, axis=1, args=(pair,))

    #     # group points into boards and get the total count for sample weighting below
    #     pair_points = self.all_point_data[self.all_point_data["in_pair"]]
    #     pair_boards = (
    #         pair_points.filter(["sync_index", "port", "point_id"])
    #         .groupby(["sync_index", "port"])
    #         .agg("count")
    #         .rename({"point_id": "point_count"}, axis=1)
    #         .query("point_count >=6")  # a requirement of the stereocalibration function
    #         .reset_index()
    #         .query(f"port == {pair[0]}")  # will be the same..only need one copy
    #         .drop("port", axis=1)
    #     )

    #     # configure random sampling. If you have too few boards, then only take what you have
    #     board_count = pair_boards.shape[0]
    #     sample_size = min(board_count, boards_sampled)

    #     if sample_size > 0:
    #         logger.info(f"Assembling {sample_size} shared boards for pair {pair}")
    #         # bias toward selecting boards with more overlapping points
    #         sample_weight = pair_boards["point_count"] ** 2

    #         # get the randomly selected subset
    #         selected_boards = pair_boards.sample(n=sample_size, weights=sample_weight, random_state=random_state)

    #         selected_pair_points = pair_points.merge(selected_boards, "right", "sync_index")
    #     else:
    #         logger.info(f"For pair {pair} there are no shared boards")
    #         selected_pair_points = None

    #     return selected_pair_points

    def stereo_calibrate_all(self, boards_sampled=10):
        """Iterates across all camera pairs. Intrinsic parameters are pulled
        from camera and combined with obj and img points for each pair.
        """
        logger.info("Deleting previous stereocalibrations from config")
        # clear out the previous stereocalibrations
        for key in self.config.copy().keys():
            if key[0:6] == "stereo":
                del self.config[key]

        logger.info(f"Beginning stereocalibration of pairs {self.pairs}")
        for pair in self.pairs:
            error, rotation, translation = self.stereo_calibrate(pair, boards_sampled)

            if error is not None:
                # only store data if there was sufficient stereopair coverage to get
                # a good calibration

                # toml dumps arrays as strings, so needs to be converted to list
                rotation = rotation.tolist()
                translation = translation.tolist()

                config_key = "stereo_" + str(pair[0]) + "_" + str(pair[1])
                self.config[config_key] = {}
                self.config[config_key]["rotation"] = rotation
                self.config[config_key]["translation"] = translation
                self.config[config_key]["RMSE"] = error

        logger.info("Direct stereocalibration complete for all pairs for which data is available")
        logger.info(f"Saving stereo-pair extrinsic data to {self.config_path}")
        with open(self.config_path, "w") as f:
            rtoml.dump(self.config, f)

    def stereo_calibrate(self, pair, boards_sampled=10):
        stereocalibration_flags = cv2.CALIB_FIX_INTRINSIC
        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 40, 0.001)

        paired_point_data = self.get_stereopair_data(pair, boards_sampled)

        if paired_point_data is not None:
            img_locs_A, obj_locs_A = self.get_stereocal_inputs(pair[0], paired_point_data)
            img_locs_B, obj_locs_B = self.get_stereocal_inputs(pair[1], paired_point_data)

            camera_matrix_A = self.config["cam_" + str(pair[0])]["matrix"]
            camera_matrix_B = self.config["cam_" + str(pair[1])]["matrix"]
            camera_matrix_A = np.array(camera_matrix_A, dtype=float)
            camera_matrix_B = np.array(camera_matrix_B, dtype=float)

            distortion_A = self.config["cam_" + str(pair[0])]["distortions"]
            distortion_B = self.config["cam_" + str(pair[1])]["distortions"]
            distortion_A = np.array(distortion_A, dtype=float)
            distortion_B = np.array(distortion_B, dtype=float)

            (
                ret,
                camera_matrix_1,
                distortion_1,
                camera_matrix_2,
                distortion_2,
                rotation,
                translation,
                essential,
                fundamental,
            ) = cv2.stereoCalibrate(
                obj_locs_A,
                img_locs_A,
                img_locs_B,
                camera_matrix_A,
                distortion_A,
                camera_matrix_B,
                distortion_B,
                # imageSize does not matter.
                # from OpenCV: "Size of the image used only to initialize the camera intrinsic matrices."
                imageSize=None,
                criteria=criteria,
                flags=stereocalibration_flags,
            )

            logger.info(f"RMSE of reprojection for pair {pair} is {ret}")

        else:
            logger.info(f"No stereocalibration produced for pair {pair}")
            ret = None
            rotation = None
            translation = None

        return ret, rotation, translation

    def get_stereocal_inputs(self, port, point_data):
        port_point_data = point_data.query(f"port == {port}")

        sync_indices = port_point_data["sync_index"].to_numpy().round().astype(int)
        img_loc_x = port_point_data["img_loc_x"].to_numpy().astype(np.float32)
        img_loc_y = port_point_data["img_loc_y"].to_numpy().astype(np.float32)
        obj_loc_x = port_point_data["obj_loc_x"].to_numpy().astype(np.float32)
        obj_loc_y = port_point_data["obj_loc_y"].to_numpy().astype(np.float32)
        obj_loc_z = obj_loc_x * 0

        # build the actual inputs for the calibration...
        img_x_y = np.vstack([img_loc_x, img_loc_y]).T
        board_x_y_z = np.vstack([obj_loc_x, obj_loc_y, obj_loc_z]).T

        # print(time.time())
        img_locs = []
        obj_locs = []
        for sync_index in np.unique(sync_indices):
            same_frame = sync_indices == sync_index
            img_locs.append(img_x_y[same_frame])
            obj_locs.append(board_x_y_z[same_frame])

        return img_locs, obj_locs



# a set of helper functions meant to provide visual feedback
# regarding the capture history and corner identification of
# the charuco board
from itertools import combinations

import cv2

import caliscope.logger

logger = caliscope.logger.get(__name__)


def grid_history(frame, ids, img_locs, connected_corners):
    """
    add the history of captured boards so that the user can see which ares of the camera FOV may not have data
    """

    possible_pairs = {pair for pair in combinations(ids, 2)}
    connected_pairs = connected_corners.intersection(possible_pairs)

    # build dictionary of corner positions:
    observed_corners = {}
    for crnr_id, crnr in zip(ids, img_locs):
        observed_corners[crnr_id] = (round(crnr[0]), round(crnr[1]))

    # add them to the visual representation of the grid capture history
    for pair in connected_pairs:
        point_1 = observed_corners[pair[0]]
        point_2 = observed_corners[pair[1]]

        cv2.line(frame, point_1, point_2, (255, 165, 0), 3)

    return frame
